import datetime
import importlib
import inspect
import os
import shutil

import yaml

from qibocal import calibrations
from qibocal.config import log, raise_error
from qibocal.data import Data


def load_yaml(path):
    """Load yaml file from disk."""
    with open(path) as file:
        data = yaml.safe_load(file)
    return data


class ActionParser:
    """Class for parsing and executing single actions in the runcard."""

    def __init__(self, runcard, folder, name):
        self.runcard = runcard
        self.folder = folder
        self.func = None
        self.params = None
        self.name = name
        self.path = os.path.join(self.folder, f"data/{self.name}/")

        # FIXME: dummy fix
        self.__name__ = name

    def build(self):
        """Load function from :func:`qibocal.characterization.calibrations` and check arguments"""
        if not os.path.exists(self.path):
            os.makedirs(self.path)
        # collect function from module
        self.func = getattr(calibrations, self.name)

        sig = inspect.signature(self.func)
        self.params = self.runcard["actions"][self.name]
        for param in list(sig.parameters)[2:-1]:
            if param not in self.params:
                raise_error(AttributeError, f"Missing parameter {param} in runcard.")

    def execute(self, data_format, platform, qubits):
        """Execute action and retrieve results."""
        if data_format is None:
            raise_error(ValueError, f"Cannot store data using {data_format} format.")

        results = self.func(platform, qubits, **self.params)

        for data in results:
            getattr(data, f"to_{data_format}")(self.path)


class niGSCactionParser(ActionParser):
    """ni = non interactive
    GSC = gate set characterization
    """

    def __init__(self, runcard, folder, name):
        super().__init__(runcard, folder, name)

        self.plots = []

        self.nqubits = self.runcard["actions"][self.name]["nqubits"]
        self.depths = self.runcard["actions"][self.name]["depths"]
        self.runs = self.runcard["actions"][self.name]["runs"]
        self.nshots = self.runcard["actions"][self.name]["nshots"]

        from qibocal.calibrations.niGSC.basics import noisemodels

        try:
            self.noise_params = self.runcard["actions"][self.name]["noise_params"]
        except KeyError:
            self.noise_params = None
        try:
            self.noise_model = getattr(
                noisemodels, self.runcard["actions"][self.name]["noise_model"]
            )(*self.noise_params)
        except:
            self.noise_model = None

    def load_plot(self):
        """Helper method to import the plotting function."""
        from qibocal.calibrations.niGSC.basics.plot import plot_qq

        self.plots.append((f"{self.name} protocol", plot_qq))

    def build(self):
        """Load appropirate module to run the experiment."""
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        self.module = importlib.import_module(f"qibocal.calibrations.niGSC.{self.name}")

    def execute(self, data_format, platform):
        """Executes a non-interactive gate set characterication using only the wanted
        module name.

        1. Build the circuit factory.
        2. Build the experiment object with the circuit factory.
        3. Execute the circuits generated by the circuit factory and store the wanted results.
        4. Post process the data.
        5. Store the experiment, if needed the circuit factory, and the aggregated data.

        Args:
            data_format (_type_): _description_
            platform (_type_): _description_
        """

        # Initiate the factory and the experiment.
        factory = self.module.ModuleFactory(
            self.nqubits, self.depths * self.runs, qubits=self.runcard["qubits"]
        )
        experiment = self.module.ModuleExperiment(
            factory, nshots=self.nshots, noise_model=self.noise_model
        )
        # Execute the circuits in the experiment.
        experiment.perform(experiment.execute)
        # Run the row by row postprocessing.
        self.module.post_processing_sequential(experiment)
        # Run aggregational tasks along with fitting.
        # This will return a data frame, store it right away.
        self.module.get_aggregational_data(experiment).to_pickle(
            f"{self.path}/fit_plot.pkl"
        )
        # Store the experiment.
        experiment.save(self.path)


class ActionBuilder:
    """Class for parsing and executing runcards.
    Args:
        runcard (path): path containing the runcard.
        folder (path): path for the output folder.
        force (bool): option to overwrite the output folder if it exists already.
    """

    def __init__(self, runcard, folder=None, force=False):
        path, self.folder = self._generate_output_folder(folder, force)
        self.runcard = load_yaml(runcard)
        # Qibolab default backend if not provided in runcard.
        backend_name = self.runcard.get("backend", "qibolab")
        platform_name = self.runcard.get("platform", "dummy")
        platform_runcard = self.runcard.get("runcard", None)
        self.backend, self.platform = self._allocate_backend(
            backend_name, platform_name, path, platform_runcard
        )
        if self.platform is not None:
            self.qubits = {
                q: self.platform.qubits[q]
                for q in self.runcard["qubits"]
                if q in self.platform.qubits
            }
        else:
            self.qubits = self.runcard.get("qubits")
        self.format = self.runcard["format"]

        # Saving runcard
        shutil.copy(runcard, f"{path}/runcard.yml")
        self.save_meta(path, self.folder)

    @staticmethod
    def _generate_output_folder(folder, force):
        """Static method for generating the output folder.
        Args:
            folder (path): path for the output folder. If None it will be created a folder automatically
            force (bool): option to overwrite the output folder if it exists already.
        """
        if folder is None:
            import getpass

            e = datetime.datetime.now()
            user = getpass.getuser().replace(".", "-")
            date = e.strftime("%Y-%m-%d")
            folder = f"{date}-{'000'}-{user}"
            num = 0
            while os.path.exists(folder):
                log.info(f"Directory {folder} already exists.")
                num += 1
                folder = f"{date}-{str(num).rjust(3, '0')}-{user}"
                log.info(f"Trying to create directory {folder}")
        elif os.path.exists(folder) and not force:
            raise_error(RuntimeError, f"Directory {folder} already exists.")
        elif os.path.exists(folder) and force:
            log.warning(f"Deleting previous directory {folder}.")
            shutil.rmtree(os.path.join(os.getcwd(), folder))

        path = os.path.join(os.getcwd(), folder)
        log.info(f"Creating directory {folder}.")
        os.makedirs(path)
        return path, folder

    def _allocate_backend(self, backend_name, platform_name, path, platform_runcard):
        """Allocate the platform using Qibolab."""
        from qibo.backends import GlobalBackend, set_backend

        if backend_name == "qibolab":
            if platform_runcard is None:
                from qibolab.paths import qibolab_folder

                original_runcard = qibolab_folder / "runcards" / f"{platform_name}.yml"
            else:
                original_runcard = platform_runcard
            # copy of the original runcard that will stay unmodified
            shutil.copy(original_runcard, f"{path}/platform.yml")
            # copy of the original runcard that will be modified during calibration
            updated_runcard = f"{self.folder}/new_platform.yml"
            shutil.copy(original_runcard, updated_runcard)
            # allocate backend with updated_runcard
            set_backend(
                backend=backend_name, platform=platform_name, runcard=updated_runcard
            )
            backend = GlobalBackend()
            return backend, backend.platform
        else:
            set_backend(backend=backend_name, platform=platform_name)
            backend = GlobalBackend()
            return backend, None

    def save_meta(self, path, folder):
        import qibocal

        e = datetime.datetime.now(datetime.timezone.utc)
        meta = {}
        meta["title"] = folder
        meta["backend"] = str(self.backend)
        meta["platform"] = str(self.backend.platform)
        meta["date"] = e.strftime("%Y-%m-%d")
        meta["start-time"] = e.strftime("%H:%M:%S")
        meta["end-time"] = e.strftime("%H:%M:%S")
        meta["versions"] = self.backend.versions  # pylint: disable=E1101
        meta["versions"]["qibocal"] = qibocal.__version__

        with open(f"{path}/meta.yml", "w") as file:
            yaml.dump(meta, file)

    def execute(self):
        """Method to execute sequentially all the actions in the runcard."""
        if self.platform is not None:
            self.platform.connect()
            self.platform.setup()
            self.platform.start()

        actions = []
        for action in self.runcard["actions"]:
            actions.append(action)
            try:
                parser = niGSCactionParser(self.runcard, self.folder, action)
                parser.build()
                parser.execute(self.format, self.platform)
            # TODO: find a better way to choose between the two parsers
            except (ModuleNotFoundError, KeyError):
                parser = ActionParser(self.runcard, self.folder, action)
                parser.build()
                parser.execute(self.format, self.platform, self.qubits)
                for qubit in self.qubits:
                    if self.platform is not None:
                        self.update_platform_runcard(qubit, action)
            self.dump_report(actions)

        if self.platform is not None:
            self.platform.stop()
            self.platform.disconnect()

    def update_platform_runcard(self, qubit, routine):
        try:
            data_fit = Data.load_data(self.folder, "data", routine, self.format, "fits")
            data_fit.df = data_fit.df[data_fit.df["qubit"] == qubit]
        except FileNotFoundError:
            return None

        params = data_fit.df.to_dict("index")[qubit if qubit == 0 else qubit - 1]
        settings = load_yaml(f"{self.folder}/new_platform.yml")
        for param in params:
            if param in list(self.qubits[qubit].__annotations__.keys()):
                setattr(self.qubits[qubit], param, params[param])
                settings["characterization"]["single_qubit"][qubit][param] = int(
                    data_fit.get_values(param)
                )

        with open(f"{self.folder}/new_platform.yml", "w") as file:
            yaml.dump(
                settings, file, sort_keys=False, indent=4, default_flow_style=None
            )

    def dump_report(self, actions=None):
        from qibocal.web.report import create_report

        # update end time
        meta = load_yaml(f"{self.folder}/meta.yml")
        e = datetime.datetime.now(datetime.timezone.utc)
        meta["end-time"] = e.strftime("%H:%M:%S")
        with open(f"{self.folder}/meta.yml", "w") as file:
            yaml.dump(meta, file)

        create_report(self.folder, actions)


class ReportBuilder:
    """Parses routines and plots to report and live plotting page.

    Args:
        path (str): Path to the data folder to generate report for.
        actions (list): List of action to be included in the report. Default is `None`
                        which corresponds to including all the actions in the qq runcard.
    """

    def __init__(self, path, actions=None):
        self.path = path
        self.metadata = load_yaml(os.path.join(path, "meta.yml"))

        # find proper path title
        base, self.title = os.path.join(os.getcwd(), path), ""
        while self.title in ("", "."):
            base, self.title = os.path.split(base)

        self.runcard = load_yaml(os.path.join(path, "runcard.yml"))
        self.format = self.runcard.get("format")
        self.qubits = self.runcard.get("qubits")

        # create calibration routine objects
        # (could be incorporated to :meth:`qibocal.cli.builders.ActionBuilder._build_single_action`)
        self.routines = []
        if actions is None:
            actions = self.runcard.get("actions")

        for action in actions:
            if hasattr(calibrations, action):
                routine = getattr(calibrations, action)
            elif hasattr(calibrations.niGSC, action):
                routine = niGSCactionParser(self.runcard, self.path, action)
                routine.load_plot()
            else:
                raise_error(ValueError, f"Undefined action {action} in report.")

            if not hasattr(routine, "plots"):
                routine.plots = []
            self.routines.append(routine)

    def get_routine_name(self, routine):
        """Prettify routine's name for report headers."""
        return routine.__name__.replace("_", " ").title()

    def get_figure(self, routine, method, qubit):
        """Get html figure for report.

        Args:
            routine (Callable): Calibration method.
            method (Callable): Plot method.
            qubit (int): Qubit id.
        """
        import tempfile

        figures, _ = method(self.path, routine.__name__, qubit, self.format)
        with tempfile.NamedTemporaryFile(delete=False) as temp:
            for figure in figures:
                figure.write_html(temp.name, include_plotlyjs=False, full_html=False)
                fightml = temp.read().decode("utf-8")
        return fightml

    def get_live_figure(self, routine, method, qubit):
        """Get url to dash page for live plotting.

        This url is used by :meth:`qibocal.web.app.get_graph`.

        Args:
            routine (Callable): Calibration method.
            method (Callable): Plot method.
            qubit (int): Qubit id.
        """
        return os.path.join(
            method.__name__,
            self.path,
            routine.__name__,
            str(qubit),
            self.format,
        )



"""Testing fitting functions"""

import logging

import numpy as np
import pytest

from qibocal.config import log
from qibocal.data import DataUnits
from qibocal.fitting.methods import (
    drag_tuning_fit,
    flipping_fit,
    lorentzian_fit,
    rabi_fit,
    ramsey_fit,
    res_spectroscopy_flux_fit,
    t1_fit,
)
from qibocal.fitting.utils import (
    cos,
    exp,
    flipping,
    freq_r_mathieu,
    freq_r_transmon,
    line,
    lorenzian,
    rabi,
    ramsey,
)


@pytest.mark.parametrize("name", [None, "test"])
@pytest.mark.parametrize(
    "qubit, fluxline, num_params",
    [
        (1, 1, 6),
        (1, 1, 7),
        (1, 2, 2),
    ],
)
def test_res_spectroscopy_flux_fit(name, qubit, fluxline, num_params, caplog):
    """Test the *res_spectrocopy_flux_fit* function"""
    x = np.linspace(-0.01, 0.03, 100)
    if num_params == 6:
        p0 = 0.01
        p1 = 41
        p2 = 0.17
        p3 = 0.75
        p4 = 78847979
        p5 = 7651970152
        params = [p0, p1, p2, p3, p4, p5]
        noisy_flux = freq_r_transmon(x, *params) + np.random.randn(100) * 1e-3
        params_fit = [p5, p4]
        labels = [
            "curr_sp",
            "xi",
            "d",
            "f_q/f_rh",
            "g",
            "f_rh",
        ]
    elif num_params == 7:
        p0 = 7651970152
        p1 = 78847979
        p2 = 0.01
        p3 = 41
        p4 = 0.17
        p5 = 2.7e8
        p6 = 1.77e10
        params = [p0, p1, p2, p3, p4, p5, p6]
        noisy_flux = freq_r_mathieu(x, *params) + np.random.randn(100) * 1e-3
        params_fit = [p0, p1, p5, p6]
        labels = [
            "f_rh",
            "g",
            "curr_sp",
            "xi",
            "d",
            "Ec",
            "Ej",
        ]
    else:
        p0 = -4366377
        p1 = 7655179288
        params = [p0, p1]
        noisy_flux = line(x, p0, p1) + np.random.randn(100) * 1e-3
        params_fit = []
        labels = [
            "popt0",
            "popt1",
        ]

    data = DataUnits(quantities={"frequency": "Hz", "current": "A"})

    mydict = {"frequency[Hz]": noisy_flux, "current[A]": x}

    data.load_data_from_dict(mydict)

    fit = res_spectroscopy_flux_fit(
        data, "current[A]", "frequency[Hz]", qubit, fluxline, params_fit
    )

    for j in range(num_params):
        np.testing.assert_allclose(fit.get_values(labels[j])[0], params[j], rtol=0.1)

    x = [0]
    noisy_flux = [0]
    data = DataUnits(quantities={"frequency": "Hz", "current": "A"})
    mydict = {"frequency[Hz]": noisy_flux, "current[A]": x}

    data.load_data_from_dict(mydict)

    fit = res_spectroscopy_flux_fit(
        data, "current[A]", "frequency[Hz]", qubit, fluxline, params_fit
    )
    assert "The fitting was not successful" in caplog.text


@pytest.mark.parametrize("name", [None, "test"])
@pytest.mark.parametrize(
    "label, resonator_type, amplitude_sign",
    [
        ("readout_frequency", "3D", 1),
        ("readout_frequency", "2D", -1),
        ("readout_frequency_shifted", "3D", 1),
        ("readout_frequency_shifted", "2D", -1),
        ("drive_frequency", "3D", -1),
        ("drive_frequency", "2D", 1),
    ],
)
def test_lorentzian_fit(name, label, resonator_type, amplitude_sign, caplog):
    """Test the *lorentzian_fit* function"""
    amplitude = 1 * amplitude_sign
    center = 2
    sigma = 3
    offset = 4

    x = np.linspace(center - 10, center + 10, 100)
    noisy_lorentzian = (
        lorenzian(x, amplitude, center, sigma, offset)
        + amplitude * np.random.randn(100) * 1e-3
    )

    data = DataUnits(quantities={"frequency": "Hz"}, options=["qubit", "iteration"])

    mydict = {
        "frequency[Hz]": x,
        "MSR[V]": noisy_lorentzian,
        "qubit": [0] * len(x),
        "iteration": [0] * len(x),
    }

    data.load_data_from_dict(mydict)

    fit = lorentzian_fit(
        data,
        "frequency[Hz]",
        "MSR[V]",
        [0],
        resonator_type,
        labels=[label, "peak_voltage"],
        fit_file_name=name,
    )
    # Given the couople (amplitude, sigma) as a solution of lorentzian_fit method
    # also (-amplitude,-sigma) is a possible solution.
    np.testing.assert_allclose(
        abs(fit.get_values("popt0")[0]), abs(amplitude), rtol=0.1
    )
    np.testing.assert_allclose(fit.get_values("popt1")[0], center, rtol=0.1)
    np.testing.assert_allclose(abs(fit.get_values("popt2")[0]), abs(sigma), rtol=0.1)
    np.testing.assert_allclose(fit.get_values("popt3")[0], offset, rtol=0.1)
    np.testing.assert_allclose(fit.get_values(label)[0], center, rtol=0.1)
    # Dummy fit
    x = [0]
    noisy_lorentzian = [0]
    data = DataUnits(quantities={"frequency": "Hz"}, options=["qubit", "iteration"])
    mydict = {
        "frequency[Hz]": x,
        "MSR[V]": noisy_lorentzian,
        "qubit": 0,
        "iteration": 0,
    }

    data.load_data_from_dict(mydict)

    fit = lorentzian_fit(
        data,
        "frequency[Hz]",
        "MSR[V]",
        [0],
        resonator_type,
        labels=[label, "peak_voltage"],
        fit_file_name=name,
    )
    assert "lorentzian_fit: the fitting was not successful" in caplog.text


@pytest.mark.parametrize(
    "label, unit, resonator_type, amplitude_sign",
    [
        ("duration", "ns", "3D", 1),
        ("duration", "ns", "2D", -1),
        ("gain", "dimensionless", "3D", 1),
        ("gain", "dimensionless", "2D", -1),
        ("amplitude", "dimensionless", "3D", 1),
        ("amplitude", "dimensionless", "2D", -1),
    ],
)
def test_rabi_fit(label, unit, resonator_type, amplitude_sign, caplog):
    """Test the *rabi_fit* function"""
    p0 = 4
    p1 = 1 * amplitude_sign
    p2 = 1
    p3 = 2
    p4 = 1 / 5 * 1e-6

    samples = 100
    x = np.linspace(0, 1 / p2, samples)
    noisy_rabi = rabi(x, p0, p1, p2, p3, p4) + p1 * np.random.randn(samples) * 1e-3

    data = DataUnits(quantities={label: unit}, options=["qubit", "iteration"])

    mydict = {
        f"{label}[{unit}]": x,
        "MSR[V]": noisy_rabi,
        "qubit": [0] * len(x),
        "iteration": [0] * len(x),
    }

    data.load_data_from_dict(mydict)

    fit = rabi_fit(
        data,
        f"{label}[{unit}]",
        "MSR[V]",
        [0],
        resonator_type,
        labels=[f"pi_pulse_{label}", "pi_pulse_peak_voltage"],
    )
    fit_p0 = fit.get_values("popt0")[0]
    fit_p1 = fit.get_values("popt1")[0]
    fit_p2 = fit.get_values("popt2")[0]
    fit_p3 = fit.get_values("popt3")[0]
    fit_p4 = fit.get_values("popt4")[0]
    y_real = p1 * np.sin(2 * np.pi * x * p2 + p3)
    y_fit = fit_p1 * np.sin(2 * np.pi * x * fit_p2 + fit_p3)
    for i in range(len(x)):
        assert abs(y_real[i] - y_fit[i]) < 0.1
    # Dummy fit
    x = [0]
    noisy_rabi = [0]

    data = DataUnits(quantities={label: unit}, options=["qubit", "iteration"])

    mydict = {
        f"{label}[{unit}]": x,
        "MSR[V]": noisy_rabi,
        "qubit": [0] * len(x),
        "iteration": [0] * len(x),
    }

    data.load_data_from_dict(mydict)

    fit = rabi_fit(
        data,
        f"{label}[{unit}]",
        "MSR[V]",
        [0],
        resonator_type,
        labels=[f"pi_pulse_{label}", "pi_pulse_peak_voltage"],
    )
    assert "rabi_fit: the fitting was not succesful" in caplog.text


@pytest.mark.parametrize(
    "resonator_type, amplitude_sign",
    [
        ("3D", 1),
        ("2D", -1),
    ],
)
def test_ramsey_fit(resonator_type, amplitude_sign, caplog):
    """Test the *ramsey_fit* function"""
    p0 = 4
    p1 = 1 * amplitude_sign
    p2 = 1
    p3 = 2
    p4 = 1 / 5 * 1e-9
    qubit_freqs = [4]
    sampling_rate = 10
    offset_freq = 1
    samples = 100
    x = np.linspace(0, 2 * np.pi / p2, samples)
    noisy_ramsey = ramsey(x, p0, p1, p2, p3, p4) + p1 * np.random.randn(samples) * 1e-3

    data = DataUnits(quantities={"wait": "ns"}, options=["qubit", "iteration"])

    mydict = {
        "wait[ns]": x,
        "MSR[V]": noisy_ramsey,
        "qubit": [0] * len(x),
        "iteration": [0] * len(x),
    }

    data.load_data_from_dict(mydict)

    fit = ramsey_fit(
        data,
        "wait[ns]",
        "MSR[V]",
        [0],
        resonator_type,
        qubit_freqs,
        sampling_rate,
        offset_freq,
        labels=["delta_frequency", "corrected_qubit_frequency", "t2"],
    )
    fit_p0 = fit.get_values("popt0")[0]
    fit_p1 = fit.get_values("popt1")[0]
    fit_p2 = fit.get_values("popt2")[0]
    fit_p3 = fit.get_values("popt3")[0]
    fit_p4 = fit.get_values("popt4")[0]

    y_real = ramsey(x, p0, p1, p2, p3, p4)
    y_fit = ramsey(x, fit_p0, fit_p1, fit_p2, fit_p3, fit_p4)
    for i in range(len(x)):
        assert abs(y_real[i] - y_fit[i]) < 0.1
    # Dummy fit
    x = [0]
    noisy_ramsey = [0]

    data = DataUnits(quantities={"wait": "ns"}, options=["qubit", "iteration"])

    mydict = {
        "wait[ns]": x,
        "MSR[V]": noisy_ramsey,
        "qubit": [0] * len(x),
        "iteration": [0] * len(x),
    }

    data.load_data_from_dict(mydict)

    fit = ramsey_fit(
        data,
        "wait[ns]",
        "MSR[V]",
        [0],
        resonator_type,
        qubit_freqs,
        sampling_rate,
        offset_freq,
        labels=["delta_frequency", "corrected_qubit_frequency", "t2"],
    )
    assert "ramsey_fit: the fitting was not succesful" in caplog.text


@pytest.mark.parametrize(
    "label, resonator_type, amplitude_sign",
    [
        ("t1", "3D", 1),
        ("t1", "2D", -1),
    ],
)
def test_t1_fit(label, resonator_type, amplitude_sign, caplog):
    """Test the *t1_fit* function"""
    p0 = 0
    p1 = 1 * amplitude_sign
    p2 = 1 / 5

    x = np.linspace(0, 10, 100)
    noisy_t1 = exp(x, p0, p1, p2) + np.random.randn(100) * 1e-4

    data = DataUnits(quantities={"wait": "ns"}, options=["qubit", "iteration"])

    mydict = {
        "wait[ns]": x,
        "MSR[V]": noisy_t1,
        "qubit": [0] * len(x),
        "iteration": [0] * len(x),
    }

    data.load_data_from_dict(mydict)

    fit = t1_fit(data, "wait[ns]", "MSR[V]", [0], resonator_type, labels=[label])

    fit_p = [fit.get_values(f"popt{i}")[0] for i in range(3)]
    fit_t1 = exp(x, *fit_p)
    for i in range(len(x)):
        assert abs(fit_t1[i] - noisy_t1[i]) < 0.2
    # Dummy fit
    x = [0]
    noisy_t1 = [0]
    data = DataUnits(quantities={"wait": "ns"}, options=["qubit", "iteration"])
    mydict = {
        "wait[ns]": x,
        "MSR[V]": noisy_t1,
        "qubit": [0] * len(x),
        "iteration": [0] * len(x),
    }

    data.load_data_from_dict(mydict)

    fit = t1_fit(data, "wait[ns]", "MSR[V]", [0], resonator_type, labels=[label])
    assert "t1_fit: the fitting was not succesful" in caplog.text


@pytest.mark.parametrize(
    "label, resonator_type, amplitude_sign",
    [
        ("amplitude_correction_factor", "3D", 1),
        ("amplitude_correction_factor", "2D", -1),
    ],
)
def test_flipping_fit(label, resonator_type, amplitude_sign, caplog):
    """Test the *flipping_fit* function"""
    p0 = 0.0001 * amplitude_sign
    p1 = 1
    p2 = 17 * amplitude_sign
    p3 = 3

    pi_pulse_amplitudes = [5]

    x = np.linspace(0, 10, 100)
    noisy_flip = flipping(x, p0, p1, p2, p3) + p0 * np.random.randn(100) * 1e-4

    data = DataUnits(
        quantities={"flips": "dimensionless"}, options=["qubit", "iteration"]
    )

    mydict = {
        "flips[dimensionless]": x,
        "MSR[V]": noisy_flip,
        "qubit": [0] * len(x),
        "iteration": [0] * len(x),
    }

    data.load_data_from_dict(mydict)

    fit = flipping_fit(
        data,
        "flips[dimensionless]",
        "MSR[V]",
        [0],
        resonator_type,
        pi_pulse_amplitudes,
        labels=[label, "corrected_amplitude"],
    )
    fit_p = [fit.get_values(f"popt{i}")[0] for i in range(4)]
    fit_flip = flipping(x, *fit_p)
    for i in range(len(x)):
        assert abs(fit_flip[i] - noisy_flip[i]) < 0.2
    # Dummy fit
    x = [0, 0]
    noisy_flip = [0, 0]
    data = DataUnits(
        quantities={"flips": "dimensionless"}, options=["qubit", "iteration"]
    )

    mydict = {
        "flips[dimensionless]": x,
        "MSR[V]": noisy_flip,
        "qubit": [0] * len(x),
        "iteration": [0] * len(x),
    }

    data.load_data_from_dict(mydict)

    fit = flipping_fit(
        data,
        "flips[dimensionless]",
        "MSR[V]",
        [0],
        resonator_type,
        pi_pulse_amplitudes,
        labels=[label, "corrected_amplitude"],
    )
    assert "flipping_fit: the fitting was not succesful" in caplog.text


@pytest.mark.parametrize(
    "label",
    [
        ("optimal_beta_param"),
    ],
)
def test_drag_tunning_fit(label, caplog):
    """Test the *drag_tunning_fit* function"""
    p0 = 0
    p1 = 0.1
    p2 = 4
    p3 = 0.3

    x = np.linspace(0, 10, 100)
    noisy_drag = cos(x, p0, p1, p2, p3) + p1 * np.random.randn(100) * 1e-6

    data = DataUnits(
        quantities={"beta_param": "dimensionless"}, options=["qubit", "iteration"]
    )

    mydict = {
        "beta_param[dimensionless]": x,
        "MSR[V]": noisy_drag,
        "qubit": [0] * len(x),
        "iteration": [0] * len(x),
    }

    data.load_data_from_dict(mydict)

    fit = drag_tuning_fit(
        data, "beta_param[dimensionless]", "MSR[V]", [0], labels=label
    )
    fit_p = [fit.get_values(f"popt{i}")[0] for i in range(4)]
    fit_drag = flipping(x, *fit_p)
    MSQE = 0
    for i in range(len(x)):
        MSQE += abs(fit_drag[i] - noisy_drag[i])
    assert MSQE / len(x) < 0.1
    # Dummy fit
    x = [0, 0]
    noisy_drag = [0, 0]
    data = DataUnits(
        quantities={"beta_param": "dimensionless"}, options=["qubit", "iteration"]
    )

    mydict = {
        "beta_param[dimensionless]": x,
        "MSR[V]": noisy_drag,
        "qubit": [0] * len(x),
        "iteration": [0] * len(x),
    }

    data.load_data_from_dict(mydict)

    fit = drag_tuning_fit(
        data, "beta_param[dimensionless]", "MSR[V]", [0], labels=label
    )
    assert "drag_tuning_fit: the fitting was not succesful" in caplog.text



import os
import pathlib

from jinja2 import Environment, FileSystemLoader

from qibocal import __version__
from qibocal.cli.builders import ReportBuilder


def create_report(path, actions=None):
    """Creates an HTML report for the data in the given path."""
    filepath = pathlib.Path(__file__)

    with open(os.path.join(filepath.with_name("static"), "styles.css")) as file:
        css_styles = f"<style>\n{file.read()}\n</style>"

    report = ReportBuilder(path, actions)
    env = Environment(loader=FileSystemLoader(filepath.with_name("templates")))
    template = env.get_template("template.html")
    html = template.render(
        is_static=True,
        css_styles=css_styles,
        version=__version__,
        reports=[report],
    )

    with open(os.path.join(path, "index.html"), "w") as file:
        file.write(html)



import os

import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from qibocal.data import Data, DataUnits
from qibocal.fitting.utils import freq_r_mathieu, freq_r_transmon, line, lorenzian
from qibocal.plots.utils import get_color, get_data_subfolders


# Resonator and qubit spectroscopies
def frequency_msr_phase(folder, routine, qubit, format):
    figures = []

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "MSR (V)",
            "phase (rad)",
        ),
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    fitting_report = ""
    for subfolder in subfolders:
        try:
            data_fast = DataUnits.load_data(
                folder, subfolder, routine, format, "fast_sweep_data"
            )
            data_fast.df = data_fast.df[data_fast.df["qubit"] == qubit]
        except:
            data_fast = DataUnits(
                quantities={"frequency": "Hz"}, options=["qubit", "iteration"]
            )
        try:
            data_precision = DataUnits.load_data(
                folder, subfolder, routine, format, f"precision_sweep_data"
            )
            data_precision.df = data_precision.df[data_precision.df["qubit"] == qubit]
        except:
            data_precision = DataUnits(
                quantities={"frequency": "Hz"}, options=["qubit", "iteration"]
            )
        try:
            data_fit = Data.load_data(folder, subfolder, routine, format, f"fits")
            data_fit.df = data_fit.df[data_fit.df["qubit"] == qubit]
        except:
            data_fit = Data(
                quantities=[
                    "popt0",
                    "popt1",
                    "popt2",
                    "popt3",
                    "label1",
                    "label2",
                    "label3",
                    "qubit",
                ]
            )

        for i, label, data in list(
            zip((0, 1), ("Fast", "Precision"), (data_fast, data_precision))
        ):
            iterations = data.df["iteration"].unique()
            frequencies = data.df["frequency"].pint.to("Hz").pint.magnitude.unique()
            if len(iterations) > 1:
                opacity = 0.3
            else:
                opacity = 1
            for iteration in iterations:
                iteration_data = data.df[data.df["iteration"] == iteration]
                fig.add_trace(
                    go.Scatter(
                        x=iteration_data["frequency"].pint.to("Hz").pint.magnitude,
                        y=iteration_data["MSR"].pint.to("uV").pint.magnitude,
                        marker_color=get_color(2 * report_n + i),
                        opacity=opacity,
                        name=f"q{qubit}/r{report_n}: {label}",
                        showlegend=not bool(iteration),
                        legendgroup=f"q{qubit}/r{report_n}: {label}",
                    ),
                    row=1,
                    col=1,
                )
                fig.add_trace(
                    go.Scatter(
                        x=iteration_data["frequency"].pint.to("Hz").pint.magnitude,
                        y=iteration_data["phase"].pint.to("rad").pint.magnitude,
                        marker_color=get_color(2 * report_n + i),
                        opacity=opacity,
                        showlegend=False,
                        legendgroup=f"q{qubit}/r{report_n}: {label}",
                    ),
                    row=1,
                    col=2,
                )
            if len(iterations) > 1:
                fig.add_trace(
                    go.Scatter(
                        x=frequencies,
                        y=data.df.groupby("frequency")["MSR"]
                        .mean()
                        .pint.to("uV")
                        .pint.magnitude,
                        marker_color=get_color(2 * report_n + i),
                        name=f"q{qubit}/r{report_n}: {label} Average",
                        showlegend=True,
                        legendgroup=f"q{qubit}/r{report_n}: {label} Average",
                    ),
                    row=1,
                    col=1,
                )
                fig.add_trace(
                    go.Scatter(
                        x=frequencies,
                        y=data.df.groupby("frequency")["phase"]
                        .mean()
                        .pint.to("rad")
                        .pint.magnitude,
                        marker_color=get_color(2 * report_n + i),
                        showlegend=False,
                        legendgroup=f"q{qubit}/r{report_n}: {label} Average",
                    ),
                    row=1,
                    col=2,
                )

        if len(data_fast) > 0 and (qubit in data_fit.df["qubit"].values):
            freqrange = np.linspace(
                min(data_fast.get_values("frequency", "Hz")),
                max(data_fast.get_values("frequency", "Hz")),
                2 * len(data_fast),
            )
            params = data_fit.df[data_fit.df["qubit"] == qubit].to_dict(
                orient="records"
            )[0]
            fig.add_trace(
                go.Scatter(
                    x=freqrange,
                    y=lorenzian(
                        freqrange,
                        data_fit.get_values("popt0"),
                        data_fit.get_values("popt1"),
                        data_fit.get_values("popt2"),
                        data_fit.get_values("popt3"),
                    ),
                    name=f"q{qubit}/r{report_n} Fit",
                    line=go.scatter.Line(dash="dot"),
                    marker_color=get_color(4 * report_n + 2),
                ),
                row=1,
                col=1,
            )

            for param, value in params.items():
                if "freq" in param:
                    fitting_report = fitting_report + (
                        f"q{qubit}/r{report_n} | {param}: {value:,.0f} Hz.<br>"
                    )
                elif "voltage" in param:
                    fitting_report = fitting_report + (
                        f"q{qubit}/r{report_n} | {param}: {value:,.0f} uV.<br>"
                    )
            fitting_report += "<br>"
        report_n += 1

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Frequency (Hz)",
        yaxis_title="MSR (uV)",
        xaxis2_title="Frequency (Hz)",
        yaxis2_title="Phase (rad)",
    )

    figures.append(fig)

    return figures, fitting_report


# Punchout
def frequency_attenuation_msr_phase(folder, routine, qubit, format):
    figures = []
    fitting_report = "No fitting data"

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)

    fig = make_subplots(
        rows=len(subfolders),
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
        subplot_titles=(
            "Normalised MSR",
            "phase (rad)",
        ),
    )

    report_n = 0
    for subfolder in subfolders:
        try:
            data = DataUnits.load_data(folder, subfolder, routine, format, f"data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = DataUnits(
                name=f"data",
                quantities={"frequency": "Hz", "attenuation": "dB"},
                options=["qubit", "iteration"],
            )

        iterations = data.df["iteration"].unique()
        frequencies = data.df["frequency"].pint.to("Hz").pint.magnitude.unique()
        attenuations = data.df["attenuation"].pint.to("dB").pint.magnitude.unique()
        averaged_data = (
            data.df.drop(columns=["qubit", "iteration"])
            .groupby(["frequency", "attenuation"], as_index=False)
            .mean()
        )

        def norm(x):
            x_mags = x.pint.to("V").pint.magnitude
            return (x_mags - np.min(x_mags)) / (np.max(x_mags) - np.min(x_mags))

        normalised_data = averaged_data.groupby(["attenuation"], as_index=False)[
            ["MSR"]
        ].transform(norm)

        fig.add_trace(
            go.Heatmap(
                x=averaged_data["frequency"].pint.to("Hz").pint.magnitude,
                y=averaged_data["attenuation"].pint.to("dB").pint.magnitude,
                z=normalised_data["MSR"],
                # z=averaged_data["MSR"].pint.to("V").pint.magnitude,
                colorbar_x=0.46,
            ),
            row=1 + report_n,
            col=1,
        )
        fig.update_xaxes(
            title_text=f"q{qubit}/r{report_n}: Frequency (Hz)", row=1 + report_n, col=1
        )
        fig.update_yaxes(title_text="Attenuation (dB)", row=1 + report_n, col=1)
        fig.add_trace(
            go.Heatmap(
                x=averaged_data["frequency"].pint.to("Hz").pint.magnitude,
                y=averaged_data["attenuation"].pint.to("dB").pint.magnitude,
                z=averaged_data["phase"].pint.to("rad").pint.magnitude,
                colorbar_x=1.01,
            ),
            row=1 + report_n,
            col=2,
        )
        fig.update_xaxes(
            title_text=f"q{qubit}/r{report_n}: Frequency (Hz)", row=1 + report_n, col=2
        )
        fig.update_yaxes(title_text="Attenuation (dB)", row=1 + report_n, col=2)
        fig.update_layout(
            showlegend=False,
            uirevision="0",  # ``uirevision`` allows zooming while live plotting
        )
        report_n += 1
    if report_n > 1:
        fig.update_traces(showscale=False)

    figures.append(fig)

    return figures, fitting_report


# Resonator and qubit spectroscopies
def frequency_attenuation_msr_phase_cut(folder, routine, qubit, format):
    figures = []
    fitting_report = "No fitting data"

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "MSR (V)",
            "phase (rad)",
        ),
    )

    subfolders = get_data_subfolders(folder)
    report_n = 0
    for subfolder in subfolders:
        try:
            data = DataUnits.load_data(folder, subfolder, routine, format, f"data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = DataUnits(
                name=f"data",
                quantities={"frequency": "Hz", "attenuation": "dB"},
                options=["qubit", "iteration"],
            )
        attenuations = data.df["attenuation"].unique()
        middle_attenuation = attenuations[len(attenuations) // 2]
        data.df = data.df[data.df["qubit"] == qubit][
            data.df["attenuation"] == middle_attenuation
        ].drop(columns=["attenuation"])

        iterations = data.df["iteration"].unique()
        frequencies = data.df["frequency"].pint.to("Hz").pint.magnitude.unique()
        if len(iterations) > 1:
            opacity = 0.3
        else:
            opacity = 1
        for iteration in iterations:
            iteration_data = data.df[data.df["iteration"] == iteration]
            fig.add_trace(
                go.Scatter(
                    x=iteration_data["frequency"].pint.to("Hz").pint.magnitude,
                    y=iteration_data["MSR"].pint.to("uV").pint.magnitude,
                    marker_color=get_color(report_n),
                    opacity=opacity,
                    name=f"q{qubit}/r{report_n} Attenuation: {middle_attenuation.to('dB').magnitude} dB",
                    showlegend=not bool(iteration),
                    legendgroup=f"q{qubit}/r{report_n}",
                ),
                row=1,
                col=1,
            )
            fig.add_trace(
                go.Scatter(
                    x=iteration_data["frequency"].pint.to("Hz").pint.magnitude,
                    y=iteration_data["phase"].pint.to("rad").pint.magnitude,
                    marker_color=get_color(report_n),
                    opacity=opacity,
                    showlegend=False,
                    legendgroup=f"q{qubit}/r{report_n}",
                ),
                row=1,
                col=2,
            )
        if len(iterations) > 1:
            fig.add_trace(
                go.Scatter(
                    x=frequencies,
                    y=data.df.groupby("frequency")["MSR"]  # pylint: disable=E1101
                    .mean()
                    .pint.to("uV")
                    .pint.magnitude,  # pylint: disable=E1101
                    marker_color=get_color(report_n),
                    name=f"q{qubit}/r{report_n}: Average",
                    showlegend=True,
                    legendgroup=f"q{qubit}/r{report_n}: Average",
                ),
                row=1,
                col=1,
            )
            fig.add_trace(
                go.Scatter(
                    x=frequencies,
                    y=data.df.groupby("frequency")["phase"]  # pylint: disable=E1101
                    .mean()
                    .pint.to("rad")
                    .pint.magnitude,  # pylint: disable=E1101
                    marker_color=get_color(report_n),
                    showlegend=False,
                    legendgroup=f"q{qubit}/r{report_n}: Average",
                ),
                row=1,
                col=2,
            )

        report_n += 1

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Frequency (Hz)",
        yaxis_title="MSR (uV)",
        xaxis2_title="Frequency (Hz)",
        yaxis2_title="Phase (rad)",
    )

    figures.append(fig)

    return figures, fitting_report


# Resonator spectroscopy flux
def frequency_flux_msr_phase(folder, routine, qubit, format):
    figures = []
    fitting_report = "No fitting data"

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)

    report_n = 0
    for subfolder in subfolders:
        try:
            data = DataUnits.load_data(folder, subfolder, routine, format, f"data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = DataUnits(
                name=f"data",
                quantities={"frequency": "Hz", "bias": "V"},
                options=["qubit", "fluxline", "iteration"],
            )

        iterations = data.df["iteration"].unique()
        fluxlines = data.df["fluxline"].unique()
        frequencies = data.df["frequency"].pint.to("Hz").pint.magnitude.unique()
        biass = data.df["bias"].pint.to("V").pint.magnitude.unique()

        if len(fluxlines) > 1:
            fig = make_subplots(
                rows=len(subfolders),
                cols=len(fluxlines),
                horizontal_spacing=0.1,
                vertical_spacing=0.1,
                subplot_titles=tuple(
                    [f"MSR [V] - fluxline {fluxline}" for fluxline in fluxlines]
                ),
            )

            for fluxline_n, fluxline in enumerate(fluxlines):
                fluxline_df = data.df[data.df["fluxline"] == fluxline]
                fluxline_df = (
                    fluxline_df.drop(columns=["qubit", "fluxline", "iteration"])
                    .groupby(["frequency", "bias"], as_index=False)
                    .mean()
                )

                fig.add_trace(
                    go.Heatmap(
                        x=fluxline_df["frequency"].pint.to("Hz").pint.magnitude,
                        y=fluxline_df["bias"].pint.to("V").pint.magnitude,
                        z=fluxline_df["MSR"].pint.to("V").pint.magnitude,
                        showscale=False,
                    ),
                    row=1 + report_n,
                    col=1 + fluxline_n,
                )
                fig.update_xaxes(
                    title_text=f"q{qubit}/r{report_n}: Frequency (GHz)",
                    row=1 + report_n,
                    col=1 + fluxline_n,
                )

        elif len(fluxlines) == 1:
            fig = make_subplots(
                rows=len(subfolders),
                cols=2,
                horizontal_spacing=0.1,
                vertical_spacing=0.1,
                subplot_titles=(
                    f"MSR [V] - fluxline {fluxlines[0]}",
                    f"Phase [rad] - fluxline {fluxlines[0]}",
                ),
            )

            fluxline_df = data.df[data.df["fluxline"] == fluxlines[0]]
            fluxline_df = (
                fluxline_df.drop(columns=["qubit", "fluxline", "iteration"])
                .groupby(["frequency", "bias"], as_index=False)
                .mean()
            )

            fig.add_trace(
                go.Heatmap(
                    x=fluxline_df["frequency"].pint.to("Hz").pint.magnitude,
                    y=fluxline_df["bias"].pint.to("V").pint.magnitude,
                    z=fluxline_df["MSR"].pint.to("V").pint.magnitude,
                    colorbar_x=0.46,
                ),
                row=1 + report_n,
                col=1,
            )
            fig.update_xaxes(
                title_text=f"q{qubit}/r{report_n}: Frequency (Hz)",
                row=1 + report_n,
                col=1,
            )

            fig.add_trace(
                go.Heatmap(
                    x=fluxline_df["frequency"].pint.to("Hz").pint.magnitude,
                    y=fluxline_df["bias"].pint.to("V").pint.magnitude,
                    z=fluxline_df["phase"].pint.to("rad").pint.magnitude,
                    colorbar_x=1.01,
                ),
                row=1 + report_n,
                col=2,
            )
            fig.update_xaxes(
                title_text=f"q{qubit}/r{report_n}: Frequency (Hz)",
                row=1 + report_n,
                col=2,
            )
            fig.update_yaxes(title_text="Current (A)", row=1 + report_n, col=2)

        fig.update_yaxes(title_text="Current (A)", row=1 + report_n, col=1)
        fig.update_layout(
            showlegend=False,
            uirevision="0",  # ``uirevision`` allows zooming while live plotting
        )

        report_n += 1
    if report_n > 1:
        fig.update_traces(showscale=False)

    figures.append(fig)

    return figures, fitting_report


# Dispersive shift
def dispersive_frequency_msr_phase(folder, routine, qubit, format):
    figures = []

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "MSR (V)",
            "phase (rad)",
        ),
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    fitting_report = ""
    for subfolder in subfolders:
        try:
            data_0 = DataUnits.load_data(folder, subfolder, routine, format, f"data_0")
            data_0.df = data_0.df[data_0.df["qubit"] == qubit]
        except:
            data_0 = DataUnits(
                name=f"data_0",
                quantities={"frequency": "Hz"},
                options=["qubit", "shifted", "iteration"],
            )
        try:
            data_1 = DataUnits.load_data(folder, subfolder, routine, format, f"data_1")
            data_1.df = data_1.df[data_1.df["qubit"] == qubit]
        except:
            data_1 = DataUnits(
                name=f"data_1",
                quantities={"frequency": "Hz"},
                options=["qubit", "shifted", "iteration"],
            )

        try:
            fit_data_0 = Data.load_data(
                folder, subfolder, routine, format, f"fit_data_0"
            )
        except:
            fit_data_0 = Data(
                quantities=[
                    "popt0",
                    "popt1",
                    "popt2",
                    "popt3",
                    "label1",
                    "label2",
                    "label3",
                ]
            )

        try:
            fit_data_1 = Data.load_data(
                folder, subfolder, routine, format, f"fit_data_1"
            )
        except:
            fit_data_1 = Data(
                quantities=[
                    "popt0",
                    "popt1",
                    "popt2",
                    "popt3",
                    "label1",
                    "label2",
                    "label3",
                ]
            )
        resonator_freqs = {}
        for i, label, data, data_fit in list(
            zip(
                (0, 1),
                ("Spectroscopy", "Shifted spectroscopy"),
                (data_0, data_1),
                (fit_data_0, fit_data_1),
            )
        ):
            iterations = data.df["iteration"].unique()
            frequencies = data.df["frequency"].pint.to("Hz").pint.magnitude.unique()
            if len(iterations) > 1:
                opacity = 0.3
            else:
                opacity = 1
            for iteration in iterations:
                iteration_data = data.df[data.df["iteration"] == iteration]
                fig.add_trace(
                    go.Scatter(
                        x=iteration_data["frequency"].pint.to("Hz").pint.magnitude,
                        y=iteration_data["MSR"].pint.to("uV").pint.magnitude,
                        marker_color=get_color(2 * report_n + i),
                        opacity=opacity,
                        name=f"q{qubit}/r{report_n}: {label}",
                        showlegend=not bool(iteration),
                        legendgroup=f"q{qubit}/r{report_n}: {label}",
                    ),
                    row=1,
                    col=1,
                )
                fig.add_trace(
                    go.Scatter(
                        x=iteration_data["frequency"].pint.to("Hz").pint.magnitude,
                        y=iteration_data["phase"].pint.to("rad").pint.magnitude,
                        marker_color=get_color(2 * report_n + i),
                        opacity=opacity,
                        showlegend=False,
                        legendgroup=f"q{qubit}/r{report_n}: {label}",
                    ),
                    row=1,
                    col=2,
                )
            if len(iterations) > 1:
                fig.add_trace(
                    go.Scatter(
                        x=frequencies,
                        y=data.df.groupby("frequency")["MSR"]
                        .mean()
                        .pint.to("uV")
                        .pint.magnitude,
                        marker_color=get_color(2 * report_n + i),
                        name=f"q{qubit}/r{report_n}: {label} Average",
                        showlegend=True,
                        legendgroup=f"q{qubit}/r{report_n}: {label} Average",
                    ),
                    row=1,
                    col=1,
                )
                fig.add_trace(
                    go.Scatter(
                        x=frequencies,
                        y=data.df.groupby("frequency")["phase"]
                        .mean()
                        .pint.to("rad")
                        .pint.magnitude,
                        marker_color=get_color(2 * report_n + i),
                        showlegend=False,
                        legendgroup=f"q{qubit}/r{report_n}: {label} Average",
                    ),
                    row=1,
                    col=2,
                )

            if len(data) > 0 and (qubit in data_fit.df["qubit"].values):
                freqrange = np.linspace(
                    min(data.get_values("frequency", "Hz")),
                    max(data.get_values("frequency", "Hz")),
                    2 * len(data),
                )
                params = data_fit.df[data_fit.df["qubit"] == qubit].to_dict(
                    orient="records"
                )[0]
                fig.add_trace(
                    go.Scatter(
                        x=freqrange,
                        y=lorenzian(
                            freqrange,
                            data_fit.get_values("popt0"),
                            data_fit.get_values("popt1"),
                            data_fit.get_values("popt2"),
                            data_fit.get_values("popt3"),
                        ),
                        name=f"q{qubit}/r{report_n}: {label} Fit",
                        line=go.scatter.Line(dash="dot"),
                        marker_color=get_color(3 * report_n + i),
                    ),
                    row=1,
                    col=1,
                )

                if "readout_frequency" in params:
                    resonator_freqs[label] = params["readout_frequency"]

                for param, value in params.items():
                    if "freq" in param:
                        fitting_report = fitting_report + (
                            f"q{qubit}/r{report_n} | {label} {param}: {value:,.0f} Hz.<br>"
                        )
                    elif "voltage" in param:
                        fitting_report = fitting_report + (
                            f"q{qubit}/r{report_n} | {label} {param}: {value:,.0f} uV.<br>"
                        )
        if ("Spectroscopy" in resonator_freqs) and (
            "Shifted spectroscopy" in resonator_freqs
        ):
            frequency_shift = (
                resonator_freqs["Shifted spectroscopy"]
                - resonator_freqs["Spectroscopy"]
            )
            fitting_report = fitting_report + (
                f"q{qubit}/r{report_n} | Frequency shift: {frequency_shift:,.0f} Hz.<br>"
            )
        fitting_report += "<br>"
        report_n += 1

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Frequency (Hz)",
        yaxis_title="MSR (uV)",
        xaxis2_title="Frequency (Hz)",
        yaxis2_title="Phase (rad)",
    )

    figures.append(fig)

    return figures, fitting_report


def frequency_attenuation(folder, routine, qubit, format):
    """Plot of the experimental data for the flux resonator flux spectroscopy and its corresponding fit.
        Args:
        folder (str): Folder where the data files with the experimental and fit data are.
        routine (str): Routine name (resonator_flux_sample_matrix)
        qubit (int): qubit coupled to the resonator for which we want to plot the data.
        format (str): format of the data files.

    Returns:
        figures (Figure): Array of figures associated to data.

    """

    figures = []
    fitting_report = "No fitting data"

    try:
        data = DataUnits.load_data(folder, "data", routine, format, f"data_q{qubit}")
    except:
        data = DataUnits(quantities={"frequency": "Hz", "attenuation": "dB"})

    try:
        data1 = DataUnits.load_data(
            folder, "data", routine, format, f"results_q{qubit}"
        )
    except:
        data1 = DataUnits(
            quantities={"snr": "dimensionless", "frequency": "Hz", "attenuation": "dB"}
        )

    fig = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        x_title="Frequency (GHz)",
        y_title="Attenuation (db)",
    )
    fig.add_trace(
        go.Scatter(
            x=data.get_values("frequency", "GHz"),
            y=data.get_values("attenuation", "dB"),
            name="Punchout",
        ),
        row=1,
        col=1,
    )
    if len(data1) > 0:
        opt_f = data1.get_values("frequency", "GHz")[0]
        opt_att = data1.get_values("attenuation", "dB")[0]
        opt_snr = data1.get_values("snr", "dimensionless")[0]
        fig.add_annotation(
            dict(
                font=dict(color="black", size=12),
                x=0,
                y=-0.30,
                showarrow=False,
                text=f"Best response found at frequency {round(opt_f,len(str(opt_f))-3)} GHz <br> for attenuation value of {opt_att} dB with snr {opt_snr :.3e}.\n",
                textangle=0,
                xanchor="left",
                xref="paper",
                yref="paper",
            )
        )
    fig.update_layout(
        margin=dict(l=60, r=20, t=20, b=130),
        autosize=False,
        width=500,
        height=500,
        uirevision="0",
    )

    figures.append(fig)

    return figures, fitting_report


def frequency_bias_flux(folder, routine, qubit, format):
    """Plot of the experimental data of the punchout.
        Args:
        folder (str): Folder where the data files with the experimental and fit data are.
        routine (str): Routine name (resonator_flux_sample_matrix)
        qubit (int): qubit coupled to the resonator for which we want to plot the data.
        format (str): format of the data files.

    Returns:
        figures [Figure]: Array of figures associated to data.

    """

    figures = []
    fitting_report = "No fitting data"

    fluxes = []
    fluxes_fit = []
    for i in range(5):  # FIXME: 5 is hardcoded
        file1 = f"{folder}/data/{routine}/data_q{qubit}_f{i}.csv"
        file2 = f"{folder}/data/{routine}/fit1_q{qubit}_f{i}.csv"
        if os.path.exists(file1):
            fluxes += [i]
        if os.path.exists(file2):
            fluxes_fit += [i]
    if len(fluxes) < 1:
        nb = 1
    else:
        nb = len(fluxes)
    fig = make_subplots(
        rows=1,
        cols=nb,
        horizontal_spacing=0.05,
        vertical_spacing=0.1,
        x_title="Current (A)",
        y_title="Frequency (GHz)",
        shared_xaxes=False,
        shared_yaxes=True,
    )
    for k, j in enumerate(fluxes):
        data_spec = DataUnits.load_data(
            folder, "data", routine, format, f"data_q{qubit}_f{j}"
        )
        fig.add_trace(
            go.Scatter(
                x=data_spec.get_values("bias", "V"),
                y=data_spec.get_values("frequency", "GHz"),
                name=f"fluxline: {j}",
                mode="markers",
            ),
            row=1,
            col=k + 1,
        )

        if j in fluxes_fit:
            try:
                data_fit = Data.load_data(
                    folder, "data", routine, format, f"fit1_q{qubit}_f{j}"
                )
            except:
                data_fit = Data(quantities=[])
            if len(data_spec) > 0 and len(data_fit) > 0:
                curr_range = np.linspace(
                    min(data_spec.get_values("bias", "V")),
                    max(data_spec.get_values("bias", "V")),
                    100,
                )
                if int(j) == int(qubit):
                    f_qs = data_fit.get_values("f_qs")[0]
                    f_rs = data_fit.get_values("f_rs")[0]
                    curr_qs = data_fit.get_values("curr_sp")[0]
                    g = data_fit.get_values("g")[0]
                    d = data_fit.get_values("d")[0]
                    xi = data_fit.get_values("xi")[0]
                    C_ii = data_fit.get_values("C_ii")[0]
                    f_offset = data_fit.get_values("f_offset")[0]
                    text_data = f"Fluxline: {j} <br> freq_r{qubit}_sp = {f_rs :.4e} Hz <br> freq_q{qubit}_sp = {f_qs :.4e} Hz <br> curr_{qubit}_sp = {curr_qs :.2e} A <br> g = {g :.2e} Hz <br> d = {d :.2e} <br> xi = {xi :.2e} 1/A <br> C_{qubit}{j} = {C_ii :.4e} Hz/A <br> f_offset_q{qubit} = {f_offset :.4e} Hz"
                    if len(data_fit.df.keys()) != 10:
                        Ec = data_fit.get_values("Ec")[0]
                        Ej = data_fit.get_values("Ej")[0]
                        text_data += f" <br> Ec = {Ec :.3e} Hz <br> Ej = {Ej :.3e} Hz"
                        freq_r_fit = freq_r_mathieu
                        params = [
                            data_fit.get_values("f_rh"),
                            data_fit.get_values("g"),
                            data_fit.get_values("curr_sp"),
                            data_fit.get_values("xi"),
                            data_fit.get_values("d"),
                            data_fit.get_values("Ec"),
                            data_fit.get_values("Ej"),
                        ]
                    else:
                        freq_r_fit = freq_r_transmon
                        params = [
                            data_fit.get_values("curr_sp"),
                            data_fit.get_values("xi"),
                            data_fit.get_values("d"),
                            data_fit.get_values("f_q/f_rh"),
                            data_fit.get_values("g"),
                            data_fit.get_values("f_rh"),
                        ]
                    y = freq_r_fit(curr_range, *params) / 10**9
                    fig.add_trace(
                        go.Scatter(
                            x=curr_range,
                            y=y,
                            name=f"Fit fluxline {j}",
                            line=go.scatter.Line(dash="dot"),
                        ),
                        row=1,
                        col=k + 1,
                    )
                    fig.add_annotation(
                        dict(
                            font=dict(color="black", size=12),
                            x=np.mean(curr_range),
                            y=-0.9,
                            showarrow=False,
                            text=text_data,
                            textangle=0,
                            xanchor="auto",
                            align="center",
                            xref=f"x{k+1}",
                            yref="paper",  # "y1",
                        )
                    )
                else:
                    fig.add_trace(
                        go.Scatter(
                            x=curr_range,
                            y=line(
                                curr_range,
                                data_fit.get_values("popt0"),
                                data_fit.get_values("popt1"),
                            )
                            / 10**9,
                            name=f"Fit fluxline {j}",
                            line=go.scatter.Line(dash="dot"),
                        ),
                        row=1,
                        col=k + 1,
                    )
                    C_ij = data_fit.get_values("popt0")[0]
                    fig.add_annotation(
                        dict(
                            font=dict(color="black", size=12),
                            x=np.mean(curr_range),
                            y=-0.9,
                            showarrow=False,
                            text=f"Fluxline: {j} <br> C_{qubit}{j} = {C_ij :.4e} Hz/A.",
                            textangle=0,
                            xanchor="center",
                            align="center",
                            xref=f"x{k+1}",
                            yref="paper",
                        )
                    )
    fig.update_layout(
        margin=dict(l=20, r=20, t=20, b=230),
        showlegend=False,
        autosize=False,
        width=500 * max(1, len(fluxes)),
        height=500,
        uirevision="0",
    )

    figures.append(fig)

    return figures, fitting_report



import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from qibocal.data import Data, DataUnits
from qibocal.fitting.utils import rabi
from qibocal.plots.utils import get_color, get_data_subfolders


# Rabi oscillations pulse length
def time_msr_phase(folder, routine, qubit, format):
    figures = []

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "MSR (V)",
            "phase (rad)",
        ),
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    fitting_report = ""
    for subfolder in subfolders:
        try:
            data = DataUnits.load_data(folder, subfolder, routine, format, "data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = DataUnits(quantities={"time": "ns"}, options=["qubit", "iteration"])

        try:
            data_fit = Data.load_data(folder, subfolder, routine, format, f"fits")
            data_fit.df = data_fit.df[data_fit.df["qubit"] == qubit]
        except:
            data_fit = Data(
                quantities=[
                    "popt0",
                    "popt1",
                    "popt2",
                    "popt3",
                    "popt4",
                    "label1",
                    "label2",
                    "qubit",
                ]
            )

        iterations = data.df["iteration"].unique()
        times = data.df["time"].pint.to("ns").pint.magnitude.unique()
        if len(iterations) > 1:
            opacity = 0.3
        else:
            opacity = 1
        for iteration in iterations:
            iteration_data = data.df[data.df["iteration"] == iteration]
            fig.add_trace(
                go.Scatter(
                    x=iteration_data["time"].pint.to("ns").pint.magnitude,
                    y=iteration_data["MSR"].pint.to("uV").pint.magnitude,
                    marker_color=get_color(report_n),
                    opacity=opacity,
                    name=f"q{qubit}/r{report_n}",
                    showlegend=not bool(iteration),
                    legendgroup=f"q{qubit}/r{report_n}",
                ),
                row=1,
                col=1,
            )
            fig.add_trace(
                go.Scatter(
                    x=iteration_data["time"].pint.to("ns").pint.magnitude,
                    y=iteration_data["phase"].pint.to("rad").pint.magnitude,
                    marker_color=get_color(report_n),
                    opacity=opacity,
                    name=f"q{qubit}/r{report_n}",
                    showlegend=False,
                    legendgroup=f"q{qubit}/r{report_n}",
                ),
                row=1,
                col=2,
            )
        if len(iterations) > 1:
            fig.add_trace(
                go.Scatter(
                    x=times,
                    y=data.df.groupby("time")["MSR"]
                    .mean()
                    .pint.to("uV")
                    .pint.magnitude,
                    marker_color=get_color(report_n),
                    name=f"q{qubit}/r{report_n}: Average",
                    showlegend=True,
                    legendgroup=f"q{qubit}/r{report_n}: Average",
                ),
                row=1,
                col=1,
            )
            fig.add_trace(
                go.Scatter(
                    x=times,
                    y=data.df.groupby("time")["phase"]  # pylint: disable=E1101
                    .mean()
                    .pint.to("rad")
                    .pint.magnitude,
                    marker_color=get_color(report_n),
                    showlegend=False,
                    legendgroup=f"q{qubit}/r{report_n}: Average",
                ),
                row=1,
                col=2,
            )

        # add fitting trace
        if len(data) > 0 and (qubit in data_fit.df["qubit"].values):
            timerange = np.linspace(
                min(data.get_values("time", "ns")),
                max(data.get_values("time", "ns")),
                2 * len(data),
            )
            params = data_fit.df[data_fit.df["qubit"] == qubit].to_dict(
                orient="records"
            )[0]
            fig.add_trace(
                go.Scatter(
                    x=timerange,
                    y=rabi(
                        timerange,
                        data_fit.get_values("popt0"),
                        data_fit.get_values("popt1"),
                        data_fit.get_values("popt2"),
                        data_fit.get_values("popt3"),
                        data_fit.get_values("popt4"),
                    ),
                    name=f"q{qubit}/r{report_n} Fit",
                    line=go.scatter.Line(dash="dot"),
                    marker_color=get_color(4 * report_n + 2),
                ),
                row=1,
                col=1,
            )

            fitting_report = (
                fitting_report
                + (
                    f"q{qubit}/r{report_n} | pi_pulse_duration: {params['pi_pulse_duration']:.2f} ns<br>"
                )
                + (
                    f"q{qubit}/r{report_n} | pi_pulse_peak_voltage: {params['pi_pulse_peak_voltage']:,.0f} uV.<br><br>"
                )
            )

        report_n += 1

    # last part
    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Time (ns)",
        yaxis_title="MSR (uV)",
        xaxis2_title="Time (ns)",
        yaxis2_title="Phase (rad)",
    )

    figures.append(fig)

    return figures, fitting_report


# Rabi oscillations pulse gain
def gain_msr_phase(folder, routine, qubit, format):
    figures = []

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "MSR (V)",
            "phase (rad)",
        ),
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    fitting_report = ""
    for subfolder in subfolders:
        try:
            data = DataUnits.load_data(folder, subfolder, routine, format, f"data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = DataUnits(
                quantities={"gain", "dimensionless"}, options=["qubit", "iteration"]
            )

        try:
            data_fit = Data.load_data(folder, subfolder, routine, format, f"fits")
            data_fit.df = data_fit.df[data_fit.df["qubit"] == qubit]
        except:
            data_fit = Data(
                quantities=[
                    "popt0",
                    "popt1",
                    "popt2",
                    "popt3",
                    "popt4",
                    "label1",
                    "label2",
                    "qubit",
                ]
            )

        iterations = data.df["iteration"].unique()
        gains = data.df["gain"].pint.to("dimensionless").pint.magnitude.unique()
        if len(iterations) > 1:
            opacity = 0.3
        else:
            opacity = 1
        for iteration in iterations:
            iteration_data = data.df[data.df["iteration"] == iteration]
            fig.add_trace(
                go.Scatter(
                    x=iteration_data["gain"].pint.to("dimensionless").pint.magnitude,
                    y=iteration_data["MSR"].pint.to("uV").pint.magnitude,
                    marker_color=get_color(report_n),
                    opacity=opacity,
                    name=f"q{qubit}/r{report_n}",
                    showlegend=not bool(iteration),
                    legendgroup=f"q{qubit}/r{report_n}",
                ),
                row=1,
                col=1,
            )
            fig.add_trace(
                go.Scatter(
                    x=iteration_data["gain"].pint.to("dimensionless").pint.magnitude,
                    y=iteration_data["phase"].pint.to("rad").pint.magnitude,
                    marker_color=get_color(report_n),
                    opacity=opacity,
                    name=f"q{qubit}/r{report_n}",
                    showlegend=False,
                    legendgroup=f"q{qubit}/r{report_n}",
                ),
                row=1,
                col=2,
            )
        if len(iterations) > 1:
            fig.add_trace(
                go.Scatter(
                    x=gains,
                    y=data.df.groupby("gain")["MSR"]
                    .mean()
                    .pint.to("uV")
                    .pint.magnitude,
                    marker_color=get_color(report_n),
                    name=f"q{qubit}/r{report_n}: Average",
                    showlegend=True,
                    legendgroup=f"q{qubit}/r{report_n}: Average",
                ),
                row=1,
                col=1,
            )
            fig.add_trace(
                go.Scatter(
                    x=gains,
                    y=data.df.groupby("gain")["phase"]  # pylint: disable=E1101
                    .mean()
                    .pint.to("rad")
                    .pint.magnitude,
                    marker_color=get_color(report_n),
                    showlegend=False,
                    legendgroup=f"q{qubit}/r{report_n}: Average",
                ),
                row=1,
                col=2,
            )

        # add fitting trace
        if len(data) > 0 and (qubit in data_fit.df["qubit"].values):
            gainrange = np.linspace(
                min(data.get_values("gain", "dimensionless")),
                max(data.get_values("gain", "dimensionless")),
                2 * len(data),
            )
            params = data_fit.df[data_fit.df["qubit"] == qubit].to_dict(
                orient="records"
            )[0]
            fig.add_trace(
                go.Scatter(
                    x=gainrange,
                    y=rabi(
                        gainrange,
                        data_fit.get_values("popt0"),
                        data_fit.get_values("popt1"),
                        data_fit.get_values("popt2"),
                        data_fit.get_values("popt3"),
                        data_fit.get_values("popt4"),
                    ),
                    name=f"q{qubit}/r{report_n} Fit",
                    line=go.scatter.Line(dash="dot"),
                    marker_color=get_color(4 * report_n + 2),
                ),
                row=1,
                col=1,
            )

            fitting_report = (
                fitting_report
                + (
                    f"q{qubit}/r{report_n} | pi_pulse_gain: {params['pi_pulse_gain']:.3f}<br>"
                )
                + (
                    f"q{qubit}/r{report_n} | pi_pulse_peak_voltage: {params['pi_pulse_peak_voltage']:,.0f} uV.<br><br>"
                )
            )

        report_n += 1

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Gain (dimensionless)",
        yaxis_title="MSR (uV)",
        xaxis2_title="Gain (dimensionless)",
        yaxis2_title="Phase (rad)",
    )

    figures.append(fig)

    return figures, fitting_report


# Rabi oscillations pulse amplitude
def amplitude_msr_phase(folder, routine, qubit, format):
    figures = []
    fitting_report = ""

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "MSR (V)",
            "phase (rad)",
        ),
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    for subfolder in subfolders:
        try:
            data = DataUnits.load_data(folder, subfolder, routine, format, f"data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = DataUnits(
                quantities={"amplitude", "dimensionless"},
                options=["qubit", "iteration"],
            )
        try:
            data_fit = Data.load_data(folder, subfolder, routine, format, f"fits")
            data_fit.df = data_fit.df[data_fit.df["qubit"] == qubit]
        except:
            data_fit = Data(
                quantities=[
                    "popt0",
                    "popt1",
                    "popt2",
                    "popt3",
                    "popt4",
                    "label1",
                    "label2",
                    "qubit",
                ]
            )

        iterations = data.df["iteration"].unique()
        amplitudes = (
            data.df["amplitude"].pint.to("dimensionless").pint.magnitude.unique()
        )
        if len(iterations) > 1:
            opacity = 0.3
        else:
            opacity = 1
        for iteration in iterations:
            iteration_data = data.df[data.df["iteration"] == iteration]
            fig.add_trace(
                go.Scatter(
                    x=iteration_data["amplitude"]
                    .pint.to("dimensionless")
                    .pint.magnitude,
                    y=iteration_data["MSR"].pint.to("uV").pint.magnitude,
                    marker_color=get_color(report_n),
                    opacity=opacity,
                    name=f"q{qubit}/r{report_n}",
                    showlegend=not bool(iteration),
                    legendgroup=f"q{qubit}/r{report_n}",
                ),
                row=1,
                col=1,
            )
            fig.add_trace(
                go.Scatter(
                    x=iteration_data["amplitude"]
                    .pint.to("dimensionless")
                    .pint.magnitude,
                    y=iteration_data["phase"].pint.to("rad").pint.magnitude,
                    marker_color=get_color(report_n),
                    opacity=opacity,
                    name=f"q{qubit}/r{report_n}",
                    showlegend=False,
                    legendgroup=f"q{qubit}/r{report_n}",
                ),
                row=1,
                col=2,
            )
        if len(iterations) > 1:
            fig.add_trace(
                go.Scatter(
                    x=amplitudes,
                    y=data.df.groupby("amplitude")["MSR"]
                    .mean()
                    .pint.to("uV")
                    .pint.magnitude,
                    marker_color=get_color(report_n),
                    name=f"q{qubit}/r{report_n}: Average",
                    showlegend=True,
                    legendgroup=f"q{qubit}/r{report_n}: Average",
                ),
                row=1,
                col=1,
            )
            fig.add_trace(
                go.Scatter(
                    x=amplitudes,
                    y=data.df.groupby("amplitude")["phase"]  # pylint: disable=E1101
                    .mean()
                    .pint.to("rad")
                    .pint.magnitude,
                    marker_color=get_color(report_n),
                    showlegend=False,
                    legendgroup=f"q{qubit}/r{report_n}: Average",
                ),
                row=1,
                col=2,
            )

        # add fitting trace
        if len(data) > 0 and (qubit in data_fit.df["qubit"].values):
            amplituderange = np.linspace(
                min(data.get_values("amplitude", "dimensionless")),
                max(data.get_values("amplitude", "dimensionless")),
                2 * len(data),
            )
            params = data_fit.df[data_fit.df["qubit"] == qubit].to_dict(
                orient="records"
            )[0]
            fig.add_trace(
                go.Scatter(
                    x=amplituderange,
                    y=rabi(
                        amplituderange,
                        data_fit.get_values("popt0"),
                        data_fit.get_values("popt1"),
                        data_fit.get_values("popt2"),
                        data_fit.get_values("popt3"),
                        data_fit.get_values("popt4"),
                    ),
                    name=f"q{qubit}/r{report_n} Fit",
                    line=go.scatter.Line(dash="dot"),
                    marker_color="rgb(255, 130, 67)",
                ),
                row=1,
                col=1,
            )

            fitting_report = (
                fitting_report
                + (
                    f"q{qubit}/r{report_n} | pi_pulse_amplitude: {params['pi_pulse_amplitude']:.3f}<br>"
                )
                + (
                    f"q{qubit}/r{report_n} | pi_pulse_peak_voltage: {params['pi_pulse_peak_voltage']:,.0f} uV.<br><br>"
                )
            )

        report_n += 1

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Amplitude (dimensionless)",
        yaxis_title="MSR (uV)",
        xaxis2_title="Amplitude (dimensionless)",
        yaxis2_title="Phase (rad)",
    )

    figures.append(fig)

    return figures, fitting_report


# Rabi pulse length and gain
def duration_gain_msr_phase(folder, routine, qubit, format):
    figures = []
    fitting_report = "No fitting data"

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)

    fig = make_subplots(
        rows=len(subfolders),
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
        subplot_titles=(
            "MSR (V)",
            "phase (rad)",
        ),
    )

    report_n = 0
    for subfolder in subfolders:
        try:
            data = DataUnits.load_data(folder, subfolder, routine, format, f"data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = DataUnits(
                name=f"data",
                quantities={"duration": "ns", "gain": "dimensionless"},
                options=["qubit", "iteration"],
            )

        iterations = data.df["iteration"].unique()
        durations = data.df["duration"].pint.to("ns").pint.magnitude.unique()
        gains = data.df["gain"].pint.to("dimensionless").pint.magnitude.unique()
        averaged_data = (
            data.df.drop(columns=["qubit", "iteration"])
            .groupby(["duration", "gain"], as_index=False)
            .mean()
        )

        fig.add_trace(
            go.Heatmap(
                x=averaged_data["duration"].pint.to("ns").pint.magnitude,
                y=averaged_data["gain"].pint.to("dimensionless").pint.magnitude,
                z=averaged_data["MSR"].pint.to("V").pint.magnitude,
                colorbar_x=0.46,
            ),
            row=1 + report_n,
            col=1,
        )
        fig.update_xaxes(
            title_text=f"q{qubit}/r{report_n}: Duration (ns)", row=1 + report_n, col=1
        )
        fig.update_yaxes(title_text="Gain (dimensionless)", row=1 + report_n, col=1)
        fig.add_trace(
            go.Heatmap(
                x=averaged_data["duration"].pint.to("ns").pint.magnitude,
                y=averaged_data["gain"].pint.to("dimensionless").pint.magnitude,
                z=averaged_data["phase"].pint.to("rad").pint.magnitude,
                colorbar_x=1.01,
            ),
            row=1 + report_n,
            col=2,
        )
        fig.update_xaxes(
            title_text=f"q{qubit}/r{report_n}: Duration (ns)", row=1 + report_n, col=2
        )
        fig.update_yaxes(title_text="Gain (dimensionless)", row=1 + report_n, col=2)
        fig.update_layout(
            showlegend=False,
            uirevision="0",  # ``uirevision`` allows zooming while live plotting
        )
        report_n += 1
    if report_n > 1:
        fig.update_traces(showscale=False)

    figures.append(fig)

    return figures, fitting_report


# Rabi pulse length and amplitude
def duration_amplitude_msr_phase(folder, routine, qubit, format):
    figures = []
    fitting_report = "No fitting data"

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)

    fig = make_subplots(
        rows=len(subfolders),
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
        subplot_titles=(
            "MSR (V)",
            "phase (rad)",
        ),
    )

    report_n = 0
    for subfolder in subfolders:
        try:
            data = DataUnits.load_data(folder, subfolder, routine, format, f"data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = DataUnits(
                name=f"data",
                quantities={"duration": "ns", "amplitude": "dimensionless"},
                options=["qubit", "iteration"],
            )

        iterations = data.df["iteration"].unique()
        durations = data.df["duration"].pint.to("ns").pint.magnitude.unique()
        amplitudes = (
            data.df["amplitude"].pint.to("dimensionless").pint.magnitude.unique()
        )
        averaged_data = (
            data.df.drop(columns=["qubit", "iteration"])
            .groupby(["duration", "amplitude"], as_index=False)
            .mean()
        )

        fig.add_trace(
            go.Heatmap(
                x=averaged_data["duration"].pint.to("ns").pint.magnitude,
                y=averaged_data["amplitude"].pint.to("dimensionless").pint.magnitude,
                z=averaged_data["MSR"].pint.to("V").pint.magnitude,
                colorbar_x=0.46,
            ),
            row=1 + report_n,
            col=1,
        )
        fig.update_xaxes(
            title_text=f"q{qubit}/r{report_n}: Duration (ns)", row=1 + report_n, col=1
        )
        fig.update_yaxes(
            title_text="Amplitude (dimensionless)", row=1 + report_n, col=1
        )
        fig.add_trace(
            go.Heatmap(
                x=averaged_data["duration"].pint.to("ns").pint.magnitude,
                y=averaged_data["amplitude"].pint.to("dimensionless").pint.magnitude,
                z=averaged_data["phase"].pint.to("rad").pint.magnitude,
                colorbar_x=1.01,
            ),
            row=1 + report_n,
            col=2,
        )
        fig.update_xaxes(
            title_text=f"q{qubit}/r{report_n}: Duration (ns)", row=1 + report_n, col=2
        )
        fig.update_yaxes(
            title_text="Amplitude (dimensionless)", row=1 + report_n, col=2
        )
        fig.update_layout(
            showlegend=False,
            uirevision="0",  # ``uirevision`` allows zooming while live plotting
        )
        report_n += 1
    if report_n > 1:
        fig.update_traces(showscale=False)

    figures.append(fig)

    return figures, fitting_report



from __future__ import annotations

import pickle
from collections.abc import Iterable
from os.path import isfile
from typing import Callable

import numpy as np
import pandas as pd
from qibo.models import Circuit
from qibo.noise import NoiseModel

from qibocal.calibrations.niGSC.basics.utils import copy_circuit, experiment_directory
from qibocal.config import raise_error


class Experiment:
    """Experiment objects which holds an iterable circuit factory along with
    a simple data structure associated to each circuit.

    Args:
        circuitfactory (Iterable): Gives a certain amount of circuits when
            iterated over.
        nshots (int): For execution of circuit, indicates how many shots.
        data (list): If filled, ``data`` can be used to specifying parameters
                     while executing a circuit or deciding how to process results.
                     It is used to store all relevant data.
    """

    def __init__(
        self,
        circuitfactory: Iterable | None,
        data: Iterable | None = None,
        nshots: int | None = 128,
        noise_model: NoiseModel | None = None,
    ) -> None:
        """ """

        if circuitfactory is not None and not isinstance(circuitfactory, Iterable):
            raise_error(
                TypeError,
                "given circuit factory has wrong type {}, must be Iterable | None.".format(
                    type(circuitfactory)
                ),
            )
        self.circuitfactory = circuitfactory
        if data is not None and not isinstance(data, Iterable):
            raise_error(
                TypeError,
                "given data has wrong type {}, must be Iterable | None ".format(
                    type(data)
                ),
            )
        self.data = data
        if nshots is not None and not isinstance(nshots, int):
            raise_error(
                TypeError,
                "given nshots has wrong type {}, must be int | None".format(
                    type(nshots)
                ),
            )
        self.nshots = nshots
        if noise_model is not None and not isinstance(noise_model, NoiseModel):
            raise_error(
                TypeError,
                "given circuit factory has wrong type {}, must be qibo NoiseModel | None .".format(
                    type(noise_model)
                ),
            )
        self.__noise_model = noise_model
        self.name = "Abstract"

    @property
    def noise_model(self):
        return self.__noise_model

    @property
    def dataframe(self) -> pd.DataFrame:
        return pd.DataFrame(self.data)

    @classmethod
    def load(cls, path: str) -> Experiment:
        """Creates an experiment object with data and if possible with circuits.

        Args:
            path (str): The directory from where the object should be restored.

        Returns:
            Experiment: The object with data (and circuitfactory).
        """
        datapath = f"{path}experiment_data.pkl"
        circuitspath = f"{path}circuits.pkl"
        if isfile(datapath):
            with open(datapath, "rb") as f:
                data = pickle.load(f)
            if isinstance(data, pd.DataFrame):  # pragma: no cover
                data = data.to_dict("records")
            nshots = len(data[0]["samples"])
        else:
            data, nshots = None, None
        if isfile(circuitspath):
            with open(circuitspath, "rb") as f:
                circuitfactory = pickle.load(f)
        else:
            circuitfactory = None
        # Initiate an instance of the experiment class.
        obj = cls(circuitfactory, data=data, nshots=nshots)
        return obj

    def save(self, path: str | None = None) -> str:
        """Creates a path if None given and pickles relevant data from ``self.data``
        and if ``self.circuitfactory`` is a list that one too.

        Returns:
            (str): The path of stored experiment.
        """

        # Check if path to store is given, if not create one. If yes check if the last character
        # is a /, if not add it.
        if path is None:
            self.path = experiment_directory("rb")
        else:
            self.path = path if path[-1] == "/" else f"{path}/"
        # Only if the circuit factory is a list it will be stored.
        if isinstance(self.circuitfactory, list):
            with open(f"{self.path}circuits.pkl", "wb") as f:
                pickle.dump(self.circuitfactory, f)
        # And only if data is not None the data list (full of dicionaries) will be
        # stored.
        if self.data is not None:
            with open(f"{self.path}experiment_data.pkl", "wb") as f:
                pickle.dump(self.data, f)
        # It is convenient to know the path after storing, so return it.
        return self.path

    def extract(
        self, output_key: str, groupby_key: str = "", agg_type: str | Callable = ""
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """Return wanted values from ``self.data`` via the dataframe property.

        If ``groupby_key`` given, aggregate the dataframe, extract the data by which the frame was
        grouped, what was calculated given the ``agg_type`` parameter. Two arrays are returned then,
        the group values and the grouped (aggregated) data. If no ``agg_type`` given use a linear function.
        If ``groupby_key`` not given, only return the extracted data from given key.

        Args:
            output_key (str): Key name of the wanted output.
            groupby_key (str): If given, group with that key name.
            agg_type (str): If given, calcuted aggregation function on groups.

        Returns:
            Either one or two np.ndarrays. If no grouping wanted, just the data. If grouping
            wanted, the values after which where grouped and the grouped data.
        """

        # Check what parameters where given.
        if not groupby_key and not agg_type:
            # No grouping and no aggreagtion is wanted. Just return the wanted output key.
            return np.array(self.dataframe[output_key].tolist())
        elif not groupby_key and agg_type:
            # No grouping wanted, just an aggregational task on all the data.
            return self.dataframe[output_key].apply(agg_type)
        elif groupby_key and not agg_type:
            # Grouping is wanted but no aggregation, use a linear function.
            df = self.dataframe.get([output_key, groupby_key])
            grouped_df = df.groupby(groupby_key, group_keys=True).apply(lambda x: x)
            return grouped_df[groupby_key].to_numpy(), grouped_df[output_key].to_numpy()
        else:
            df = self.dataframe.get([output_key, groupby_key])
            grouped_df = df.groupby(groupby_key, group_keys=True).apply(agg_type)
            return grouped_df.index.to_numpy(), grouped_df[output_key].to_numpy()

    def prebuild(self) -> None:
        """Converts the attribute ``circuitfactory`` which is in general
        an iterable into a list.
        """
        self.circuitfactory = list(self.circuitfactory)

    def perform(self, sequential_task: Callable[[Circuit, dict], dict]) -> None:
        """Takes a given function, checks the status of attribute ``circuitfactory``
        and ``data`` and executes the sequential function row by row altering the
        ``self.data`` attribute.

        Either ``self.circuitfactory`` or ``self.data`` cannot be ``None`` and
        if not ``None`` they have to have the right length.

        Args:
            sequential_task (callable[[Circuit, dict], dict]): A function applied
                row by row alterting each datarow.
        """
        # Either the circuit factory or the data rows can be empty.
        # If ``self.data`` is not empty the actual list element is altered without
        # storing it after alternation.
        # Both ``circuit`` and ``datarow`` can be provided:
        if self.circuitfactory is not None and self.data is not None:
            for circuit, datarow in zip(self.circuitfactory, self.data):
                datarow = sequential_task(
                    copy_circuit(circuit.copy(deep=True)), datarow
                )
        # Only``datarow`` can be provided:
        elif self.circuitfactory is None and self.data is not None:
            for datarow in self.data:
                datarow = sequential_task(None, datarow)
        # Only ``circuit`` can be provided:
        elif self.circuitfactory is not None and self.data is None:
            newdata = []
            for circuit in self.circuitfactory:
                newdata.append(
                    sequential_task(copy_circuit(circuit.copy(deep=True)), {})
                )
            self.data = newdata
        else:
            raise_error(ValueError, "Both attributes circuitfactory and data are None.")

    def execute(self, circuit: Circuit, datarow: dict) -> dict:
        """Executes a circuit, returns the single shot results in a dict.

        Args:
            circuit (Circuit): Will be executed, has to return samples.
            datarow (dict): Dictionary with parameters for execution and immediate
                postprocessing information.
        """

        if self.noise_model is not None:
            circuit = self.noise_model.apply(circuit)
        samples = circuit(nshots=self.nshots).samples()
        return {"samples": samples}



from collections.abc import Iterable
from copy import deepcopy

import numpy as np
import pytest
from qibo import gates, models

from qibocal.calibrations.niGSC.basics.circuitfactory import *
from qibocal.calibrations.niGSC.basics.utils import ONEQ_GATES
from qibocal.calibrations.niGSC.standardrb import (
    ModuleFactory as SingleCliffordsInvFactory,
)
from qibocal.calibrations.niGSC.XIdrb import ModuleFactory as XIdFactory


@pytest.fixture
def factories_singlequbitgates():
    thelist = [
        SingleCliffordsFactory,
        Qibo1qGatesFactory,
        SingleCliffordsInvFactory,
        XIdFactory,
    ]
    return thelist


@pytest.fixture
def depths():
    return [0, 1, 5, 10, 30]


def abstract_factorytest(gfactory):
    # The factory is an iterable.
    assert isinstance(gfactory, Iterable)
    # The objects it produces are of the type ``models.Circuit``.
    for circuit in gfactory:
        assert isinstance(circuit, models.Circuit)


def test_abstract_factory():
    cfactory = CircuitFactory(1, [1, 2] * 3, qubits=[0])
    with pytest.raises(NotImplementedError):
        for circuit in cfactory:
            print(circuit.draw())
    cfactory = CircuitFactory(1, 3, qubits=[0])
    assert cfactory.depths == [3]


@pytest.mark.parametrize("nqubits", [1, 2, 5])
@pytest.mark.parametrize("runs", [1, 4])
@pytest.mark.parametrize("qubits", [[0], [0, 2]])
def test_general_singlequbitgates_factories(
    factories_singlequbitgates: list,
    nqubits: int,
    qubits: list,
    depths: list,
    runs: int,
) -> None:
    """Check for how random circuits are produced and if the lengths, shape
    and randomness works.

    Args:
        qubits (list): List of qubits
        depths (list): list of depths for circuits
        runs (int): How many randomly drawn cirucit for one depth value

    """
    if max(qubits) >= nqubits:
        pass
    else:
        for factory_init in factories_singlequbitgates:
            # XId factory is only defined for 1 qubit.
            if max(qubits) > 0 and factory_init == XIdFactory:
                with pytest.raises(ValueError):
                    factory = factory_init(nqubits, list(depths) * runs, qubits=qubits)
            else:
                factory = factory_init(nqubits, list(depths) * runs, qubits=qubits)
                abstract_factorytest(factory)
                # if factory.name not in ('XId', 'SingleCliffordsInv'):
                if "inv" in factory.name or "Inv" in factory.name:
                    # When checking the depth of circuits, the measurement gate and inverse gate
                    # has to be taken into account
                    additional_gates = 2
                else:
                    # When checking the depth of circuits, measurement gate has to be taken into account
                    additional_gates = 1
                for count, circuit in enumerate(factory):
                    if circuit.ngates == 1:
                        assert isinstance(circuit.queue[0], gates.measurements.M)
                    else:
                        assert (
                            circuit.ngates
                            == depths[count % len(depths)] * len(qubits)
                            + additional_gates
                        )
                        assert (
                            circuit.depth
                            == depths[count % len(depths)] + additional_gates
                        )
                    # Check the factories individual trades.
                    if factory.name in ("Qibo1qGates"):
                        for gate in circuit.queue[:-1]:
                            assert gate.__class__.__name__ in ONEQ_GATES
                    elif factory.name in ("SingleCliffords"):
                        for gate in circuit.queue[:-1]:
                            assert isinstance(gate, gates.Unitary)
                    elif factory.name in ("SingleCliffordsInv"):
                        for gate in circuit.queue[:-1]:
                            assert isinstance(gate, gates.Unitary)
                    elif factory.name in ("XId"):
                        for gate in circuit.queue[:-1]:
                            assert isinstance(gate, gates.X) or isinstance(
                                gate, gates.I
                            )
                    else:
                        raise_error(
                            ValueError,
                            "Unknown circuitfactory :{}".format(factory.name),
                        )



import os
from shutil import rmtree

from qibocal.cli._base import ActionBuilder


def test_command_niGSC():
    path_to_runcard = "src/qibocal/tests/niGSC.yml"
    test_folder = "test_and_delete/"
    builder = ActionBuilder(path_to_runcard, test_folder, force=None)
    builder.execute()
    builder.dump_report()
    paths_to_protocols = [
        "data/simulfilteredrb/",
        "data/standardrb/",
        "data/XIdrb/",
    ]
    paths_to_check = [f"{test_folder}{path}" for path in paths_to_protocols]
    for path in paths_to_check:
        assert os.path.isdir(path)
    inside = ["experiment_data.pkl", "fit_plot.pkl"]
    files_to_check = [[f"{path}{name}" for path in paths_to_check] for name in inside]
    for filename_list in files_to_check:
        for filename in filename_list:
            assert os.path.isfile(filename)
    # The circuits data is only stored with correlated rb.
    assert os.path.isfile(f"{test_folder}/data/simulfilteredrb/circuits.pkl")
    assert not os.path.isfile(f"{test_folder}/data/standardrb/circuits.pkl")
    assert not os.path.isfile(f"{test_folder}/data/XIdrb/circuits.pkl")
    assert os.path.isfile(f"{test_folder}index.html")
    assert os.path.isfile(f"{test_folder}meta.yml")
    assert os.path.isfile(f"{test_folder}runcard.yml")
    rmtree(test_folder)



from copy import deepcopy
from os import mkdir
from os.path import isdir
from typing import Union

import numpy as np
from qibo.models import Circuit

# Gates, without having to define any paramters
ONEQ_GATES = ["I", "X", "Y", "Z", "H", "S", "SDG", "T", "TDG"]


def experiment_directory(name: str):
    """Make the directory where the experiment will be stored."""
    from datetime import datetime

    overall_dir = "experiments/"
    # Check if the overall directory exists. If not create it.
    if not isdir(overall_dir):
        mkdir(overall_dir)
    # Get the current date and time.
    dt_string = datetime.now().strftime("%y%b%d_%H%M%S")
    # Every script name ``name`` gets its own directory.
    subdirectory = f"{overall_dir}{name}/"
    if not isdir(subdirectory):  # pragma: no cover
        mkdir(subdirectory)
    # Name the final directory for this experiment.
    final_directory = f"{subdirectory}experiment{dt_string}/"
    if not isdir(final_directory):  # pragma: no cover
        mkdir(final_directory)
    else:
        already_file, count = True, 1
        while already_file:
            final_directory = f"{subdirectory}experiment{dt_string}_{count}/"
            if not isdir(final_directory):
                mkdir(final_directory)
                already_file = False
            else:
                count += 1
    return final_directory


def effective_depol(error_channel, **kwargs):
    """ """
    liouvillerep = error_channel.to_pauli_liouville(normalize=True)
    d = int(np.sqrt(len(liouvillerep)))
    depolp = (np.trace(liouvillerep) - 1) / (d**2 - 1)
    return depolp


def probabilities(allsamples: Union[list, np.ndarray]) -> np.ndarray:
    """Takes the given list/array (3-dimensional) of samples and returns probabilities
    for each possible state to occure.

    The states for 4 qubits are order as follows:
    [(0, 0, 0, 0), (0, 0, 0, 1), (0, 0, 1, 0), (0, 0, 1, 1), (0, 1, 0, 0),
    (0, 1, 0, 1), (0, 1, 1, 0), (0, 1, 1, 1), (1, 0, 0, 0), (1, 0, 0, 1),
    (1, 0, 1, 0), (1, 0, 1, 1), (1, 1, 0, 0), (1, 1, 0, 1), (1, 1, 1, 0), (1, 1, 1, 1)]

    Args:
        allsamples (Union[list, np.ndarray]): The single shot samples, 3-dimensional.

    Returns:
        np.ndarray: Probability array of 2 dimension.
    """

    from itertools import product

    # Make it an array to use the shape property.
    allsamples = np.array(allsamples)
    # The array has to have three dimension.
    if len(allsamples.shape) == 2:
        allsamples = allsamples[None, ...]
    nqubits, nshots = len(allsamples[0][0]), len(allsamples[0])
    # Create all possible state vectors.
    allstates = list(product([0, 1], repeat=nqubits))
    # Iterate over all the samples and count the different states.
    probs = [
        [np.sum(np.product(samples == state, axis=1)) for state in allstates]
        for samples in allsamples
    ]
    probs = np.array(probs) / (nshots)
    return probs


def copy_circuit(circuit: Circuit) -> Circuit:
    """Truly deepcopies a given circuit by copying the gates.


    Right now, qibos copy function changes properties of the copied circuit.

    Args:
        circuit (Circuit): The circuit which is copied.

    Returns:
        (Circuit): The copied circuit
    """
    newcircuit = Circuit(circuit.nqubits)
    for gate in circuit.queue:
        newcircuit.add(deepcopy(gate))
    return newcircuit


def gate_fidelity(eff_depol: float, primitive=False) -> float:
    """Returns the average gate fidelity given the effective depolarizing parameter for single qubits.

    If primitive is True, divide by additional 1.875 as convetion in RB reporting.
    (The original reasoning was that Clifford gates are typically
    compiled with an average number of 1.875 Pi half pulses.)

    Args:
        eff_depol (float): The effective depolarizing parameter.
        primitive (bool, optional): If True, additionally divide by 1.875.

    Returns:
        float: Average gate fidelity
    """
    infidelity = (1 - eff_depol) / 2
    if primitive:
        infidelity /= 1.875
    return 1 - infidelity


def number_to_str(number: Union[int, float, complex]) -> str:
    """Converts a number into a string.

    Necessary when storing a complex number in JASON format.

    Args:
        number (int | float | complex)

    Returns:
        str: The number expressed as a string, with two floating points when
        complex or three when real.
    """
    if np.iscomplex(number):
        the_str = "{:.2f}{}{:.2f}j".format(
            np.real(number),
            "+" if np.imag(number) >= 0 else "-",
            np.abs(np.imag(number)),
        )
    else:
        the_str = "{:.3f}".format(number)
    return the_str



import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence

from qibocal import plots
from qibocal.data import DataUnits
from qibocal.decorators import plot
from qibocal.fitting.methods import rabi_fit


@plot("MSR vs Time", plots.time_msr_phase)
def rabi_pulse_length(
    platform: AbstractPlatform,
    qubits: dict,
    pulse_duration_start,
    pulse_duration_end,
    pulse_duration_step,
    software_averages=1,
    points=10,
):
    r"""
    In the Rabi experiment we apply a pulse at the frequency of the qubit and scan the drive pulse duration
    to find the drive pulse length that creates a rotation of a desired angle.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): Dict of target Qubit objects to perform the action
        pulse_duration_start (int): Initial drive pulse duration for the Rabi experiment
        pulse_duration_end (int): Maximum drive pulse duration for the Rabi experiment
        pulse_duration_step (int): Scan range step for the drive pulse duration for the Rabi experiment
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **time[ns]**: Drive pulse duration in ns
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

        - A DataUnits object with the fitted data obtained with the following keys

            - **pi_pulse_duration**: pi pulse duration
            - **pi_pulse_peak_voltage**: pi pulse's maximum voltage
            - **popt0**: offset
            - **popt1**: oscillation amplitude
            - **popt2**: frequency
            - **popt3**: phase
            - **popt4**: T2
            - **qubit**: The qubit being tested
    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create a sequence of pulses for the experiment
    sequence = PulseSequence()
    qd_pulses = {}
    ro_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_qubit_drive_pulse(qubit, start=0, duration=4)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=qd_pulses[qubit].finish
        )
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameter to sweep and its range:
    # qubit drive pulse duration time
    qd_pulse_duration_range = np.arange(
        pulse_duration_start, pulse_duration_end, pulse_duration_step
    )

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include qubit drive pulse duration time
    data = DataUnits(
        name="data", quantities={"time": "ns"}, options=["qubit", "iteration"]
    )

    # repeat the experiment as many times as defined by software_averages
    count = 0
    for iteration in range(software_averages):
        # sweep the parameter
        for duration in qd_pulse_duration_range:
            for qubit in qubits:
                qd_pulses[qubit].duration = duration
                ro_pulses[qubit].start = duration
            # save data as often as defined by points
            if count % points == 0 and count > 0:
                # save data
                yield data
                # calculate and save fit
                yield rabi_fit(
                    data,
                    x="time[ns]",
                    y="MSR[uV]",
                    qubits=qubits,
                    resonator_type=platform.resonator_type,
                    labels=[
                        "pi_pulse_duration",
                        "pi_pulse_peak_voltage",
                    ],
                )

            # execute the pulse sequence
            results = platform.execute_pulse_sequence(sequence)

            for ro_pulse in ro_pulses.values():
                # average msr, phase, i and q over the number of shots defined in the runcard
                r = results[ro_pulse.serial].to_dict()
                r.update(
                    {
                        "time[ns]": duration,
                        "qubit": ro_pulse.qubit,
                        "iteration": iteration,
                    }
                )
                data.add(r)
            count += 1
    yield data
    yield rabi_fit(
        data,
        x="time[ns]",
        y="MSR[uV]",
        qubits=qubits,
        resonator_type=platform.resonator_type,
        labels=[
            "pi_pulse_duration",
            "pi_pulse_peak_voltage",
        ],
    )


@plot("MSR vs Gain", plots.gain_msr_phase)
def rabi_pulse_gain(
    platform: AbstractPlatform,
    qubits: dict,
    pulse_gain_start,
    pulse_gain_end,
    pulse_gain_step,
    software_averages=1,
    points=10,
):
    r"""
    In the Rabi experiment we apply a pulse at the frequency of the qubit and scan the drive pulse gain
    to find the drive pulse gain that creates a rotation of a desired angle.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): Dict of target Qubit objects to perform the action
        pulse_gain_start (int): Initial drive pulse gain for the Rabi experiment
        pulse_gain_end (int): Maximum drive pulse gain for the Rabi experiment
        pulse_gain_step (int): Scan range step for the drive pulse gain for the Rabi experiment
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **gain[dimensionless]**: Drive pulse gain
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

        - A DataUnits object with the fitted data obtained with the following keys

            - **pi_pulse_duration**: pi pulse duration
            - **pi_pulse_peak_voltage**: pi pulse's maximum voltage
            - **popt0**: offset
            - **popt1**: oscillation amplitude
            - **popt2**: frequency
            - **popt3**: phase
            - **popt4**: T2
            - **qubit**: The qubit being tested
    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create a sequence of pulses for the experiment
    sequence = PulseSequence()
    qd_pulses = {}
    ro_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_RX_pulse(qubit, start=0)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=qd_pulses[qubit].finish
        )
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameter to sweep and its range:
    # qubit drive pulse gain
    qd_pulse_gain_range = np.arange(pulse_gain_start, pulse_gain_end, pulse_gain_step)

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include qubit drive pulse gain
    data = DataUnits(
        name=f"data",
        quantities={"gain": "dimensionless"},
        options=["qubit", "iteration"],
    )

    count = 0
    for iteration in range(software_averages):
        # sweep the parameter
        for gain in qd_pulse_gain_range:
            for qubit in qubits:
                platform.set_gain(qubit, gain)
            # save data as often as defined by points
            if count % points == 0 and count > 0:
                # save data
                yield data
                # calculate and save fit
                yield data
                yield rabi_fit(
                    data,
                    x="gain[dimensionless]",
                    y="MSR[uV]",
                    qubits=qubits,
                    resonator_type=platform.resonator_type,
                    labels=[
                        "pi_pulse_gain",
                        "pi_pulse_peak_voltage",
                    ],
                )

            # execute the pulse sequence
            results = platform.execute_pulse_sequence(sequence)

            for ro_pulse in ro_pulses.values():
                # average msr, phase, i and q over the number of shots defined in the runcard
                r = results[ro_pulse.serial].to_dict()
                r.update(
                    {
                        "gain[dimensionless]": gain,
                        "qubit": ro_pulse.qubit,
                        "iteration": iteration,
                    }
                )
                data.add(r)
            count += 1
    yield data
    yield rabi_fit(
        data,
        x="gain[dimensionless]",
        y="MSR[uV]",
        qubits=qubits,
        resonator_type=platform.resonator_type,
        labels=[
            "pi_pulse_gain",
            "pi_pulse_peak_voltage",
        ],
    )


@plot("MSR vs Amplitude", plots.amplitude_msr_phase)
def rabi_pulse_amplitude(
    platform: AbstractPlatform,
    qubits: dict,
    pulse_amplitude_start,
    pulse_amplitude_end,
    pulse_amplitude_step,
    software_averages=1,
    points=10,
):
    r"""
    In the Rabi experiment we apply a pulse at the frequency of the qubit and scan the drive pulse amplitude
    to find the drive pulse amplitude that creates a rotation of a desired angle.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): Dict of target Qubit objects to perform the action
        pulse_amplitude_start (int): Initial drive pulse amplitude for the Rabi experiment
        pulse_amplitude_end (int): Maximum drive pulse amplitude for the Rabi experiment
        pulse_amplitude_step (int): Scan range step for the drive pulse amplitude for the Rabi experiment
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **amplitude[dimensionless]**: Drive pulse amplitude
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

        - A DataUnits object with the fitted data obtained with the following keys

            - **pi_pulse_amplitude**: pi pulse amplitude
            - **pi_pulse_peak_voltage**: pi pulse's maximum voltage
            - **popt0**: offset
            - **popt1**: oscillation amplitude
            - **popt2**: frequency
            - **popt3**: phase
            - **popt4**: T2
            - **qubit**: The qubit being tested
    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create a sequence of pulses for the experiment
    sequence = PulseSequence()
    qd_pulses = {}
    ro_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_RX_pulse(qubit, start=0)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=qd_pulses[qubit].finish
        )
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameter to sweep and its range:
    # qubit drive pulse amplitude
    qd_pulse_amplitude_range = np.arange(
        pulse_amplitude_start, pulse_amplitude_end, pulse_amplitude_step
    )

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include qubit drive pulse amplitude
    data = DataUnits(
        name=f"data",
        quantities={"amplitude": "dimensionless"},
        options=["qubit", "iteration"],
    )

    count = 0
    for iteration in range(software_averages):
        # sweep the parameter
        for amplitude in qd_pulse_amplitude_range:
            for qubit in qubits:
                qd_pulses[qubit].amplitude = amplitude
            # save data as often as defined by points
            if count % points == 0 and count > 0:
                yield data
                # calculate and save fit
                yield rabi_fit(
                    data,
                    x="amplitude[dimensionless]",
                    y="MSR[uV]",
                    qubits=qubits,
                    resonator_type=platform.resonator_type,
                    labels=[
                        "pi_pulse_amplitude",
                        "pi_pulse_peak_voltage",
                    ],
                )

            # execute the pulse sequence
            results = platform.execute_pulse_sequence(sequence)

            for ro_pulse in ro_pulses.values():
                # average msr, phase, i and q over the number of shots defined in the runcard
                r = results[ro_pulse.serial].to_dict()
                r.update(
                    {
                        "amplitude[dimensionless]": amplitude,
                        "qubit": ro_pulse.qubit,
                        "iteration": iteration,
                    }
                )
                data.add(r)
            count += 1
    yield data
    yield rabi_fit(
        data,
        x="amplitude[dimensionless]",
        y="MSR[uV]",
        qubits=qubits,
        resonator_type=platform.resonator_type,
        labels=[
            "pi_pulse_amplitude",
            "pi_pulse_peak_voltage",
        ],
    )


@plot("MSR vs length and gain", plots.duration_gain_msr_phase)
def rabi_pulse_length_and_gain(
    platform: AbstractPlatform,
    qubits: dict,
    pulse_duration_start,
    pulse_duration_end,
    pulse_duration_step,
    pulse_gain_start,
    pulse_gain_end,
    pulse_gain_step,
    software_averages=1,
    points=10,
):
    r"""
    In the Rabi experiment we apply a pulse at the frequency of the qubit and scan the drive pulse
    combination of duration and gain to find the drive pulse amplitude that creates a rotation of a desired angle.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): Dict of target Qubit objects to perform the action
        pulse_duration_start (int): Initial drive pulse duration for the Rabi experiment
        pulse_duration_end (int): Maximum drive pulse duration for the Rabi experiment
        pulse_duration_step (int): Scan range step for the drive pulse duration for the Rabi experiment
        pulse_gain_start (int): Initial drive pulse gain for the Rabi experiment
        pulse_gain_end (int): Maximum drive pulse gain for the Rabi experiment
        pulse_gain_step (int): Scan range step for the drive pulse gain for the Rabi experiment
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **duration[ns]**: Drive pulse duration in ns
            - **gain[dimensionless]**: Drive pulse gain
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create a sequence of pulses for the experiment
    sequence = PulseSequence()
    qd_pulses = {}
    ro_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_qubit_drive_pulse(qubit, start=0, duration=4)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=qd_pulses[qubit].finish
        )
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameters to sweep and their range:
    # qubit drive pulse duration time
    qd_pulse_duration_range = np.arange(
        pulse_duration_start, pulse_duration_end, pulse_duration_step
    )
    # qubit drive pulse gain
    qd_pulse_gain_range = np.arange(pulse_gain_start, pulse_gain_end, pulse_gain_step)

    # create a DataUnits object to store the results
    # that includes qubit drive pulse duration and gain
    data = DataUnits(
        name=f"data",
        quantities={"duration": "ns", "gain": "dimensionless"},
        options=["qubit", "iteration"],
    )

    count = 0
    for iteration in range(software_averages):
        # sweep the parameters
        for duration in qd_pulse_duration_range:
            for gain in qd_pulse_gain_range:
                for qubit in qubits:
                    qd_pulses[qubit].duration = duration
                    ro_pulses[qubit].start = duration
                    platform.set_gain(qubit, gain)
                # save data as often as defined by points
                if count % points == 0 and count > 0:
                    # save data
                    yield data

                # execute the pulse sequence
                results = platform.execute_pulse_sequence(sequence)
                for ro_pulse in ro_pulses.values():
                    # average msr, phase, i and q over the number of shots defined in the runcard
                    r = results[ro_pulse.serial].to_dict()
                    r.update(
                        {
                            "duration[ns]": duration,
                            "gain[dimensionless]": gain,
                            "qubit": ro_pulse.qubit,
                            "iteration": iteration,
                        }
                    )
                    data.add(r)
                count += 1
    yield data


@plot("MSR vs length and amplitude", plots.duration_amplitude_msr_phase)
def rabi_pulse_length_and_amplitude(
    platform,
    qubits: dict,
    pulse_duration_start,
    pulse_duration_end,
    pulse_duration_step,
    pulse_amplitude_start,
    pulse_amplitude_end,
    pulse_amplitude_step,
    software_averages=1,
    points=10,
):
    r"""
    In the Rabi experiment we apply a pulse at the frequency of the qubit and scan the drive pulse
    combination of duration and amplitude to find the drive pulse amplitude that creates a rotation of a desired angle.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubit (int): Target qubit to perform the action
        pulse_duration_start (int): Initial drive pulse duration for the Rabi experiment
        pulse_duration_end (int): Maximum drive pulse duration for the Rabi experiment
        pulse_duration_step (int): Scan range step for the drive pulse duration for the Rabi experiment
        pulse_amplitude_start (int): Initial drive pulse amplitude for the Rabi experiment
        pulse_amplitude_end (int): Maximum drive pulse amplitude for the Rabi experiment
        pulse_amplitude_step (int): Scan range step for the drive pulse amplitude for the Rabi experiment
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **duration[ns]**: Drive pulse duration in ns
            - **amplitude[dimensionless]**: Drive pulse amplitude
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create a sequence of pulses for the experiment
    sequence = PulseSequence()
    qd_pulses = {}
    ro_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_qubit_drive_pulse(qubit, start=0, duration=4)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=qd_pulses[qubit].finish
        )
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameters to sweep and their range:
    # qubit drive pulse duration time
    qd_pulse_duration_range = np.arange(
        pulse_duration_start, pulse_duration_end, pulse_duration_step
    )
    # qubit drive pulse amplitude
    qd_pulse_amplitude_range = np.arange(
        pulse_amplitude_start, pulse_amplitude_end, pulse_amplitude_step
    )

    # create a DataUnits object to store the results
    # that includes qubit drive pulse duration and amplitude
    data = DataUnits(
        name=f"data",
        quantities={"duration": "ns", "amplitude": "dimensionless"},
        options=["qubit", "iteration"],
    )

    count = 0
    for iteration in range(software_averages):
        # sweep the parameters
        for duration in qd_pulse_duration_range:
            for amplitude in qd_pulse_amplitude_range:
                for qubit in qubits:
                    qd_pulses[qubit].duration = duration
                    ro_pulses[qubit].start = duration
                    qd_pulses[qubit].amplitude = amplitude
                # save data as often as defined by points
                if count % points == 0 and count > 0:
                    # save data
                    yield data

                # execute the pulse sequence
                results = platform.execute_pulse_sequence(sequence)
                for ro_pulse in ro_pulses.values():
                    # average msr, phase, i and q over the number of shots defined in the runcard
                    r = results[ro_pulse.serial].to_dict()
                    r.update(
                        {
                            "duration[ns]": duration,
                            "amplitude[dimensionless]": amplitude,
                            "qubit": ro_pulse.qubit,
                            "iteration": iteration,
                        }
                    )
                    data.add(r)
                count += 1
    yield data



import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from qibocal.data import Data, DataUnits
from qibocal.fitting.utils import exp
from qibocal.plots.utils import get_color, get_data_subfolders


# T1
def t1_time_msr(folder, routine, qubit, format):
    figures = []

    fig = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=("MSR (V)",),
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    fitting_report = ""
    for subfolder in subfolders:
        try:
            data = DataUnits.load_data(folder, subfolder, routine, format, f"data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = DataUnits(
                name=f"data",
                quantities={"wait": "ns"},
                options=["qubit", "iteration"],
            )

        try:
            data_fit = Data.load_data(folder, subfolder, routine, format, f"fits")
            data_fit.df = data_fit.df[data_fit.df["qubit"] == qubit]
        except:
            data_fit = Data(
                quantities=[
                    "popt0",
                    "popt1",
                    "popt2",
                    "label1",
                    "qubit",
                ]
            )

        iterations = data.df["iteration"].unique()
        waits = data.df["wait"].pint.to("ns").pint.magnitude.unique()

        if len(iterations) > 1:
            opacity = 0.3
        else:
            opacity = 1
        for iteration in iterations:
            iteration_data = data.df[data.df["iteration"] == iteration]
            fig.add_trace(
                go.Scatter(
                    x=iteration_data["wait"].pint.to("ns").pint.magnitude,
                    y=iteration_data["MSR"].pint.to("uV").pint.magnitude,
                    marker_color=get_color(report_n),
                    opacity=opacity,
                    name=f"q{qubit}/r{report_n}",
                    showlegend=not bool(iteration),
                    legendgroup=f"q{qubit}/r{report_n}",
                ),
                row=1,
                col=1,
            )

        if len(iterations) > 1:
            fig.add_trace(
                go.Scatter(
                    x=waits,
                    y=data.df.groupby("wait")["MSR"]
                    .mean()
                    .pint.to("uV")
                    .pint.magnitude,
                    marker_color=get_color(report_n),
                    name=f"q{qubit}/r{report_n}: Average",
                    showlegend=True,
                    legendgroup=f"q{qubit}/r{report_n}: Average",
                ),
                row=1,
                col=1,
            )

        # add fitting trace
        if len(data) > 0 and (qubit in data_fit.df["qubit"].values):
            waitrange = np.linspace(
                min(data.get_values("wait", "ns")),
                max(data.get_values("wait", "ns")),
                2 * len(data),
            )
            params = data_fit.df[data_fit.df["qubit"] == qubit].to_dict(
                orient="records"
            )[0]

            fig.add_trace(
                go.Scatter(
                    x=waitrange,
                    y=exp(
                        waitrange,
                        data_fit.get_values("popt0"),
                        data_fit.get_values("popt1"),
                        data_fit.get_values("popt2"),
                    ),
                    name=f"q{qubit}/r{report_n} Fit",
                    line=go.scatter.Line(dash="dot"),
                    marker_color=get_color(4 * report_n + 2),
                ),
                row=1,
                col=1,
            )

            fitting_report = fitting_report + (
                f"q{qubit}/r{report_n} | t1: {params['T1']:,.0f} ns.<br><br>"
            )

        report_n += 1

    # last part
    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Time (ns)",
        yaxis_title="MSR (uV)",
    )

    figures.append(fig)

    return figures, fitting_report



import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

import qibocal.calibrations.niGSC.basics.fitting as fitting_methods
from qibocal.calibrations.niGSC.basics import utils
from qibocal.calibrations.niGSC.basics.experiment import Experiment


def plot_qq(folder: str, routine: str, qubit, format):
    fitting_report = ""
    """Load the module for which the plot has to be done.


    Args:
        folder (str): The folder path where the data was stored.
        routine (str): The routine name, here the module name.
        qubit (Any): Is not used here
        format (Any): Is not used here.

    Returns:
        plotly figure: plotly graphical object figure.
    """

    import importlib

    # Load the module, something like 'standardrb'.
    module = importlib.import_module(f"qibocal.calibrations.niGSC.{routine}")
    # Load the experiment with the class method ``load``.
    experiment = module.ModuleExperiment.load(f"{folder}/data/{routine}/")
    # In this data frame the precomputed fitting parameters and other
    # parameters for fitting and plotting are stored.
    aggr_df = pd.read_pickle(f"{folder}/data/{routine}/fit_plot.pkl")
    # Build the figure/report using the responsible module.
    plotly_figure = module.build_report(experiment, aggr_df)
    return [plotly_figure], fitting_report


class Report:
    """Once initialized with the correct parameters an Report object can build
    reports to display results of a gate set characterization experiment.
    """

    def __init__(self) -> None:
        self.all_figures = []
        self.title = "Report"
        self.info_dict = {}

    def build(self):
        l = len(self.all_figures)
        if l < 3:
            divide_by = 1
        else:
            divide_by = 2
        subplot_titles = [figdict.get("subplot_title") for figdict in self.all_figures]
        fig = make_subplots(
            rows=int(l / divide_by) + l % divide_by + 1,
            cols=1 if l == 1 else divide_by,
            subplot_titles=subplot_titles,
        )
        for count, fig_dict in enumerate(self.all_figures):
            plot_list = fig_dict["figs"]
            for plot in plot_list:
                fig.add_trace(
                    plot, row=count // divide_by + 1, col=count % divide_by + 1
                )

        fig.add_annotation(
            dict(
                bordercolor="black",
                font=dict(color="black", size=16),
                x=0.0,
                y=1.0 / (int(l / divide_by) + l % divide_by + 1)
                - len(self.info_dict) * 0.005,
                showarrow=False,
                text="<br>".join(
                    [f"{key} : {value}\n" for key, value in self.info_dict.items()]
                ),
                align="left",
                textangle=0,
                yanchor="top",
                xref="paper",
                yref="paper",
            )
        )
        fig.update_xaxes(title_font_size=18, tickfont_size=16)
        fig.update_yaxes(title_font_size=18, tickfont_size=16)
        fig.update_layout(
            font_family="Averta",
            hoverlabel_font_family="Averta",
            title_text=self.title,
            title_font_size=24,
            legend_font_size=16,
            hoverlabel_font_size=16,
            showlegend=True,
            height=500 * (int(l / divide_by) + l % divide_by)
            if l > divide_by
            else 1000,
            width=1000,
        )

        return fig


def scatter_fit_fig(
    experiment: Experiment,
    df_aggr: pd.DataFrame,
    xlabel: str,
    index: str,
    fittingparam_label="popt",
):
    fig_traces = []
    dfrow = df_aggr.loc[index]
    fig_traces.append(
        go.Scatter(
            x=experiment.dataframe[xlabel],
            y=experiment.dataframe[index],
            line=dict(color="#6597aa"),
            mode="markers",
            marker={"opacity": 0.2, "symbol": "square"},
            name="runs",
        )
    )
    fig_traces.append(
        go.Scatter(
            x=dfrow[xlabel],
            y=dfrow["data"],
            line=dict(color="#aa6464"),
            mode="markers",
            name="average",
        )
    )
    x_fit = np.linspace(min(dfrow[xlabel]), max(dfrow[xlabel]), len(dfrow[xlabel]) * 20)
    if "imag" in fittingparam_label:
        y_fit = np.imag(
            getattr(fitting_methods, dfrow["fit_func"])(
                x_fit, *dfrow[fittingparam_label].values()
            )
        )
    else:
        y_fit = np.real(
            getattr(fitting_methods, dfrow["fit_func"])(
                x_fit, *dfrow[fittingparam_label].values()
            )
        )
    fig_traces.append(
        go.Scatter(
            x=x_fit,
            y=y_fit,
            name="".join(
                [
                    "{}:{} ".format(
                        key, utils.number_to_str(dfrow[fittingparam_label][key])
                    )
                    for key in dfrow[fittingparam_label]
                ]
            ),
            line=go.scatter.Line(dash="dot"),
        )
    )
    return {"figs": fig_traces, "xlabel": xlabel, "ylabel": index}



import os
from shutil import rmtree

import numpy as np
import pytest
from qibo.models import Circuit

from qibocal.calibrations.niGSC.basics.circuitfactory import Qibo1qGatesFactory
from qibocal.calibrations.niGSC.basics.experiment import *
from qibocal.calibrations.niGSC.basics.noisemodels import PauliErrorOnX


@pytest.fixture
def depths():
    return [0, 1, 5, 10, 30]


def test_Experiment_init():
    cfactory = None
    data = None
    noise_model = None
    # All None should work.
    experiment1 = Experiment(cfactory, data=data, noise_model=noise_model)
    with pytest.raises(TypeError):
        _ = Experiment(1)
        _ = Experiment(None, 1)
        _ = Experiment(None, None, True)
        _ = Experiment(None, None, None, 1)


@pytest.mark.parametrize("nqubits", [2, 3])
@pytest.mark.parametrize("runs", [1, 3])
@pytest.mark.parametrize("qubits", [[0], [0, 1]])
@pytest.mark.parametrize("nshots", [12, 23])
def test_Experiment_execute(
    nqubits: int, depths: list, runs: int, qubits: list, nshots: int
):
    cfactory1 = Qibo1qGatesFactory(nqubits, depths * runs, qubits=qubits)
    experiment1 = Experiment(cfactory1, nshots=nshots)
    experiment1.perform(experiment1.execute)
    assert experiment1.extract("samples").shape == (
        len(depths * runs),
        nshots,
        len(qubits),
    )
    noise_model = PauliErrorOnX()
    cfactory2 = Qibo1qGatesFactory(nqubits, depths * runs, qubits=qubits)
    experiment2 = Experiment(cfactory2, nshots=nshots, noise_model=noise_model)
    experiment2.perform(experiment2.execute)
    assert experiment2.extract("samples").shape == (
        len(depths * runs),
        nshots,
        len(qubits),
    )


@pytest.mark.parametrize("nqubits", [1, 2])
@pytest.mark.parametrize("runs", [1, 3])
def test_Experiment_perform(nqubits: int, depths: list, runs: int):
    def nocircuit_dict_empty(circuit: Circuit, datadict: dict):
        datadict["nshots"] = np.random.randint(1, 10)
        return datadict

    def nocirucuit_dict_filled(circuit: Circuit, datadict: dict):
        datadict["depthstr"] = "".join((*datadict.keys(), str(datadict["depth"])))
        return datadict

    def circuit_nodata(circuit: Circuit, datadict: dict):
        datadict["depth"] = circuit.depth
        return datadict

    def circuit_data(circuit: Circuit, datadict: dict):
        nshots = datadict["nshots"]
        datadict["samples"] = circuit(nshots=nshots).samples()
        return datadict

    cfactory1 = Qibo1qGatesFactory(nqubits, depths * runs)
    checkdepths = np.array(depths * runs) + 1
    checkdepthswithstring = [f"depth{d}" for d in checkdepths]
    experiment1 = Experiment(None)
    with pytest.raises(ValueError):
        experiment1.perform(circuit_nodata)
    experiment1.circuitfactory = cfactory1
    experiment1.perform(circuit_nodata)
    assert np.allclose(experiment1.dataframe.values.flatten(), checkdepths)
    experiment1.perform(nocirucuit_dict_filled)
    assert np.array_equal(experiment1.dataframe.values[:, 0].flatten(), checkdepths)
    assert np.array_equal(
        experiment1.dataframe.values[:, 1].flatten(), checkdepthswithstring
    )
    assert experiment1.dataframe.columns[0] == "depth"
    assert experiment1.dataframe.columns[1] == "depthstr"
    cfactory2 = Qibo1qGatesFactory(nqubits, depths * runs)
    experiment2 = Experiment(cfactory2)
    experiment2.perform(nocircuit_dict_empty)
    assert experiment2.dataframe.columns[0] == "nshots"
    experiment2.perform(circuit_data)
    assert experiment2.dataframe.columns[0] == "nshots"
    assert experiment2.dataframe.columns[1] == "samples"
    for item in experiment2.data:
        assert len(item["samples"]) == item["nshots"]
        assert isinstance(item["samples"], np.ndarray)


@pytest.mark.parametrize("nqubits", [2, 3])
@pytest.mark.parametrize("runs", [1, 3])
@pytest.mark.parametrize("qubits", [[0], [0, 1]])
def test_Experiment_save_load(nqubits: int, depths: list, runs: int, qubits: list):
    cfactory1 = Qibo1qGatesFactory(nqubits, depths * runs, qubits=qubits)
    experiment1 = Experiment(cfactory1)
    experiment1.perform(experiment1.execute)
    path1 = experiment1.save()
    experiment2 = Experiment.load(path1)
    for datarow1, datarow2 in zip(experiment1.data, experiment2.data):
        assert np.array_equal(datarow1["samples"], datarow2["samples"])
    assert experiment2.circuitfactory is None

    cfactory3 = Qibo1qGatesFactory(nqubits, depths * runs, qubits=qubits)
    experiment3 = Experiment(cfactory3)
    experiment3.prebuild()
    path3 = experiment3.save()
    experiment4 = Experiment.load(path3)
    for circuit3, circuit4 in zip(
        experiment3.circuitfactory, experiment4.circuitfactory
    ):
        assert np.array_equal(circuit3.unitary(), circuit4.unitary())

    cfactory5 = Qibo1qGatesFactory(nqubits, depths * runs, qubits=qubits)
    experiment5 = Experiment(cfactory5)
    experiment5.prebuild()
    path5 = experiment5.save()

    experiment6 = Experiment.load(path5)
    assert experiment6.data is None
    for circuit5, circuit6 in zip(
        experiment5.circuitfactory, experiment6.circuitfactory
    ):
        assert np.array_equal(circuit5.unitary(), circuit6.unitary())

    rmtree(path1)
    rmtree(path3)
    rmtree(path5)
    if len(os.listdir("experiments/rb")) == 0:
        rmtree("experiments/rb")
    if len(os.listdir("experiments")) == 0:
        rmtree("experiments/")


@pytest.mark.parametrize("amount_data", [50, 71])
def test_Experiment_extract(amount_data):
    def populate_experiment_data(circuit: Circuit, datarow: dict) -> dict:
        datarow["rand_uniform"] = np.random.uniform(0, 1)
        datarow["rand_normal"] = np.random.uniform(0, 1)
        datarow["rand_int"] = np.random.randint(0, 2) + 10
        return datarow

    data = [{} for _ in range(amount_data)]
    experiment1 = Experiment(None, data=data)
    experiment1.perform(populate_experiment_data)
    # No group key and not agg_type.
    randnormal_array = experiment1.extract("rand_normal")
    # No groupkey but agg_type.
    assert np.mean(randnormal_array) == experiment1.extract("rand_normal", "", "mean")
    # No agg_type but group key.
    rand_ints, rand_uniforms = experiment1.extract(
        "rand_uniform", groupby_key="rand_int"
    )
    indx_randint = np.argsort(experiment1.dataframe["rand_int"].to_numpy())
    randints_tocheck = experiment1.dataframe["rand_int"].to_numpy()[indx_randint]
    assert np.array_equal(randints_tocheck, rand_ints)
    assert len(np.unique(rand_ints)) == 2
    until = np.count_nonzero(randints_tocheck == rand_ints[0])
    a_bit_sorted_uniforms = experiment1.dataframe["rand_uniform"].to_numpy()[
        indx_randint
    ]
    each_randint_uniforms = [rand_uniforms[:until], rand_uniforms[until:]]
    each_randint_uniforms_check = [
        a_bit_sorted_uniforms[:until],
        a_bit_sorted_uniforms[until:],
    ]
    # Since they does not neccessarily have to be ordered the same way:
    for array1, array2 in zip(each_randint_uniforms, each_randint_uniforms_check):
        for element in array1:
            assert element in array2
    # Group key and agg_type.
    rand_ints, mean_uniforms = experiment1.extract("rand_uniform", "rand_int", "mean")
    assert np.isclose(mean_uniforms[0], np.mean(each_randint_uniforms_check[0]))
    assert np.isclose(mean_uniforms[1], np.mean(each_randint_uniforms_check[1]))
